/*  Eyetracking software creates and executes experiments using 
    Eyelink II hardware and libraries by SR Research

    Copyright (C) 2004  David J. Stracuzzi, University of Massachusetts

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "stdafx.h"
#include "EyeDoctor.h"
#include "DataView.h"
#include "FixationEvent.h"
#include "SelectionState.h"
#include "TrialData.h"
#include ".\dataview.h"

IMPLEMENT_DYNCREATE(CDataView, CFormView)

CDataView::CDataView()
  : CFormView(CDataView::IDD), _state(0), _offset(0)
{
	const int nCols = 7;
	const int nRows = 9;

	for(int i=0; i < nRows; ++i)
		for(int j=0; j < nCols; ++j)
			_table[i][j] = 0;
}

CDataView::~CDataView()
{}

void CDataView::DoDataExchange(CDataExchange* pDX)
{
	CFormView::DoDataExchange(pDX);
}

BEGIN_MESSAGE_MAP(CDataView, CFormView)
	ON_WM_SIZE()
	ON_WM_PAINT()
	ON_WM_VSCROLL()
END_MESSAGE_MAP()

#ifdef _DEBUG
void CDataView::AssertValid() const
{
	CScrollView::AssertValid();
}

void CDataView::Dump(CDumpContext& dc) const
{
	CScrollView::Dump(dc);
}
#endif //_DEBUG

void CDataView::OnSize(UINT nType, int cx, int cy)
{
	CSize size = GetTotalSize();
	CFormView::OnSize(nType, cx, cy);
	RedrawWindow();
}

void CDataView::UpdateRow(int row, FixationEvent *fEvt)
{
	if(fEvt) {
		const int maxLen = 20;
		EyeEvent *ePtr = fEvt->previous();
		char buf[maxLen];
		CPoint chLoc = fEvt->charLocation();
		int saccDur, i = 0;

		_itoa(fEvt->index(), buf, 10);           //construct fixation index
		while(i < maxLen && buf[i] != '\0') ++i;
		buf[i++] = ' ';
		buf[i++] = fEvt->isRightEye() ? 'R' : 'L';
		buf[i++] = '\0';

		while(ePtr) {                            //find prev fixation
			if(dynamic_cast<FixationEvent*>(ePtr))
				break;
			else ePtr = ePtr->previous();
		}                                        //compute saccade duration as difference of fixations
		if(ePtr) saccDur = fEvt->startTime() - ePtr->endTime();
		else saccDur = 0;

		if(!fEvt->isRightEye())
			_table[row][0] = -fEvt->index();
		else _table[row][0] = fEvt->index();
		_table[row][1] = chLoc.x;
		_table[row][2] = chLoc.y;
		_table[row][3] = fEvt->startTime();
		_table[row][4] = fEvt->endTime();
		_table[row][5] = fEvt->duration();
		_table[row][6] = saccDur;
	}
	else {
		const int nCols = 7;

		for(int i=0; i < nCols; ++i)
			_table[row][i] = 0;
	}
}

void CDataView::SetSelectionState(SelectionState *state)
{
	_state = state;
}

void CDataView::UpdateView()
{
	const int nRows = 9;

	if(!_state->selected().empty()) {            //if a fixation is selected
		EyeEvent *curPtr = _state->selected().back();
		int i = nRows/2-1;

		curPtr = curPtr->previous();
		while(i >= 0) {                          //write four fixations preceeding the selected
			if(curPtr) {                         //have an event
				FixationEvent *event = dynamic_cast<FixationEvent*>(curPtr);

				if(event)                        //event is a fixation, write a row
					UpdateRow(i--, event);
				curPtr = curPtr->previous();
			}
			else UpdateRow(i--, 0);
		}

		curPtr = _state->selected().back();
		i = nRows/2;
		while(i < nRows) {                       //write selected fixation and following four
			if(curPtr) {                         //have an event
				FixationEvent *event = dynamic_cast<FixationEvent*>(curPtr);

				if(event)                        //event is a fixation, write a row
					UpdateRow(i++, event);
				curPtr = curPtr->next();         //not a fixation, skip it
			}
			else UpdateRow(i++, 0);
		}
	}
	else {                                       //no selected fixation, clear all rows
		for(int i = 0; i < nRows; ++i)
			UpdateRow(i, 0);
	}

	RedrawWindow();
}

void CDataView::OnPaint()
{
	const int nColumns = 7;
	CPaintDC dc(this);                           // device context for painting
	CBrush yellow(0x00FFFF);
	CFont labelFont, *oldFont;
	CRect highlightRect, labelRect, wndRect;
	CSize size = GetTotalSize();
	COLORREF bkColor = dc.GetBkColor();
	int columnWidth;

	GetWindowRect(&wndRect);
	highlightRect.SetRect(0, 7*_rowHeight+_offset, wndRect.Width(), 8*_rowHeight+_offset);
	dc.FillRect(&highlightRect, &yellow);

	if(wndRect.Width() < size.cx)
		columnWidth = size.cx/7;
	else 
		columnWidth = wndRect.Width() / nColumns;
	labelRect.SetRect(0, _offset, columnWidth, _rowHeight+_offset);

	labelFont.CreatePointFont(100, "Courier New", &dc);
	oldFont = dc.SelectObject(&labelFont);

	dc.DrawText("Fixation\nNumber", 15, labelRect, DT_CENTER | DT_NOCLIP);
	labelRect.MoveToX(columnWidth);
	dc.DrawText("Character\nColumn", 16, labelRect, DT_CENTER | DT_NOCLIP);
	labelRect.MoveToX(2*columnWidth);
	dc.DrawText("Character\nRow", 13, labelRect, DT_CENTER | DT_NOCLIP);
	labelRect.MoveToX(3*columnWidth);
	dc.DrawText("Fixation\nStart Time", 19, labelRect, DT_CENTER | DT_NOCLIP);
	labelRect.MoveToX(4*columnWidth);
	dc.DrawText("Fixation\nEnd Time", 17, labelRect, DT_CENTER | DT_NOCLIP);
	labelRect.MoveToX(5*columnWidth);
	dc.DrawText("Fixation\nDuration", 17, labelRect, DT_CENTER | DT_NOCLIP);
	labelRect.MoveToX(6*columnWidth);
	dc.DrawText("Saccade\nDuration", 16, labelRect, DT_CENTER | DT_NOCLIP);
	labelRect.MoveToX(7*columnWidth);

	for(int row = 0; row < 9; ++row) {
		const int maxLen = 20;
		char buf[maxLen], eye = 'R';
		int i=0;
		bool isRight = true;

		labelRect.MoveToXY(0, (row+3)*_rowHeight+_offset);
		if(row == 4) dc.SetBkColor(0x00FFFF);
		else dc.SetBkColor(bkColor);

		if(_table[row][0] < 0) {
			_table[row][0] = -_table[row][0];
			eye = 'L';
		}
		else if(_table[row][0] == 0)
			eye = ' ';

		_itoa(_table[row][0], buf, 10);           //construct fixation index
		while(i < maxLen && buf[i] != '\0') ++i;
		buf[i++] = ' ';
		buf[i++] = eye;
		buf[i++] = '\0';
		dc.DrawText(buf, -1, labelRect, DT_CENTER | DT_NOCLIP | DT_SINGLELINE);

		for(int col = 1; col < 7; ++col) {
			labelRect.MoveToX(col*columnWidth);
			_itoa(_table[row][col], buf, 10);
			dc.DrawText(buf, -1, labelRect, DT_CENTER | DT_NOCLIP | DT_SINGLELINE);
		}
	}

	dc.SelectObject(oldFont);
}

void CDataView::OnVScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar)
{
	SCROLLINFO   info;
	int curpos = GetScrollPos(SB_VERT);             //Get current position of scroll box.
	int minpos, maxpos;

	GetScrollInfo(SB_VERT, &info, SIF_ALL);         //Get scrolling info
	GetScrollRange(SB_VERT, &minpos, &maxpos);      //Get min and max scroll-bar positions.
	maxpos = GetScrollLimit(SB_VERT);

	switch (nSBCode) {                              //Determine the new position of scroll box.
	case SB_TOP:
		curpos = minpos;
		break;
	case SB_BOTTOM:
		curpos = maxpos;
		break;
	case SB_ENDSCROLL:
		break;
	case SB_LINEUP:
		if (curpos > minpos)
			--curpos;
		break;
	case SB_LINEDOWN:
		if (curpos < maxpos)
			++curpos;
		break;
	case SB_PAGEUP:
   		if (curpos > minpos)
			curpos = max(minpos, curpos - (int) info.nPage);
		break;
	case SB_PAGEDOWN:
		if (curpos < maxpos)
			curpos = min(maxpos, curpos + (int) info.nPage);
		break;
	case SB_THUMBPOSITION:
		curpos = nPos;
		break;
	case SB_THUMBTRACK:
		curpos = nPos;
		break;
	}
	
	SetScrollPos(SB_VERT, curpos);                  //Set new position of scroll box.
	_offset = -curpos;
	RedrawWindow();
	CFormView::OnVScroll(nSBCode, nPos, pScrollBar);
}
