/*  Eyetracking software creates and executes experiments using 
    Eyelink II hardware and libraries by SR Research

    Copyright (C) 2004  David J. Stracuzzi, University of Massachusetts

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "StdAfx.h"
#include <algorithm>
#include <list>
#include <sstream>
#include "ExperimentData.h"
#include "ProgressDialog.h"
#include "TrialSelectionDialog.h"

ExperimentData::ExperimentData()
  : _displaySize(0, 0, 0, 0)
{}

ExperimentData::~ExperimentData()
{
	const int sz = static_cast<int>(_trial.size());
	CProgressDialog progressDialog;              //dialog for the progress bar

	progressDialog.Create(IDD_PROGRESS_DIALOG, 0);
	progressDialog.SetWindowText("Unloading Experiment");
	progressDialog.SetRange(0, sz);
	progressDialog.SetPosition(0);

	for(unsigned int i = 0; i < _trial.size(); ++i) {
		std::string message("Unloading trial ");

		progressDialog.SetPosition(i);
		message.append(_trial[i]->trialID().idString());
		progressDialog.SetMessage(message.c_str());
		delete _trial[i];
	}
}

const CRect &ExperimentData::displaySize() const
{
	return(_displaySize);
}

void ExperimentData::export(const std::string &fileName, std::ostream &changeStrm, 
							const ExportParams &param) throw(std::exception)
{
	int nSkipped = 0;
	CProgressDialog progressDialog;              //dialog for the progress bar
	std::ofstream exportStrm(fileName.c_str(), std::ios::out);

	if(!exportStrm.is_open())
		throw(std::exception("Error opening file for data export."));

	progressDialog.Create(IDD_PROGRESS_DIALOG, 0);
	progressDialog.SetWindowText(("Exporting " + fileName).c_str());
	progressDialog.SetRange(0, static_cast<int>(_trial.size()));
	progressDialog.SetPosition(0);

	int nSkipped_jhook = 0;
	int nSkipped_late= 0;
	int nSkipped_blink = 0;
	int nSkipped_longSacc = 0;
	int nTotal = 0;
	int condSkipped[300];
	for (int p=0; p<300;p++){
		condSkipped[p] = 0;
	}

	for(unsigned int i = 0; i < _trial.size(); ++i) {
		progressDialog.SetPosition(i);


		if(_trial[i]->testFlag(TRIAL_DELETED) == false &&
		  (param.exportDep || !_trial[i]->isDependent()) &&
		  (!param.excludeTens || _trial[i]->dcWithin10) &&
		  (!param.excludeFives || _trial[i]->dcWithin5) &&
		  (!param.excludeZeros || _trial[i]->dcOK) &&
		  (param.exportJHooks || !_trial[i]->testFlag(J_HOOK_DISP_CHANGE)) &&
		  (param.exportBlinkErrors || !_trial[i]->testFlag(BLINK_CHANGE_ERROR)) && 
		  (!param.excludeLongDCSaccs || !_trial[i]->hasLongSacc )
		  
		  
		  ) {
			
			std::string message("Exporting trial ");

			message.append(_trial[i]->trialID().idString());
			progressDialog.SetMessage(message.c_str());
			changeStrm<<message<<"\r\n";
			_trial[i]->export(exportStrm, changeStrm, param);
		}
		else {
			_trial[i]->setFlag(TRIAL_DELETED,true);
			changeStrm<<"Skipping trial "<<_trial[i]->trialID().idString()<<"\r\n";
			if(!param.exportJHooks && _trial[i]->testFlag(J_HOOK_DISP_CHANGE))
				nSkipped_jhook++;
			if(param.excludeZeros && !_trial[i]->dcOK ||
				param.excludeFives && !_trial[i]->dcWithin5 ||
				param.excludeTens && !_trial[i]->dcWithin10 ){
				nSkipped_late++;
				}
				if(!param.exportBlinkErrors && _trial[i]->testFlag(BLINK_CHANGE_ERROR)){
					nSkipped_blink++;
				}
			if (_trial[i]->trialID().condition() <300) 
				condSkipped[_trial[i]->trialID().condition()] += 1;
			if (_trial[i]->hasLongSacc){
				nSkipped_longSacc++;
			}

			int dd = _trial[i]->trialID().condition();
		nSkipped ++;
		}
		nTotal++;
	}
	changeStrm<<"***********************\r\n";
	for(int i = 0; i<300; i++){
		if(condSkipped[i])
			changeStrm<<"Number Skipped from condition "<<i<<": "<<condSkipped[i]<<"\r\n";
	}
	changeStrm<<"Total Number Skipped: "<<nSkipped<<"\r\n";
	changeStrm<<"Number J Hooks Skipped: "<<nSkipped_jhook<<"\r\n";
	changeStrm<<"Number Blink Change Errors Skipped: "<<nSkipped_blink<<"\r\n";
	changeStrm<<"Number Late Disp Change Skipped: "<<nSkipped_late<<"\r\n";
	changeStrm<<"Number Long Saccade Trials Skipped: "<<nSkipped_longSacc<<"\r\n";
	changeStrm<<"Total Trials considered: "<<nTotal<<"\r\n";



	exportStrm.close();
}

void ExperimentData::import(const std::string &dataFileName) throw(std::exception)
{
	CProgressDialog progressDialog;              //dialog for the progress bar
	CTrialSelectionDialog dialog;                //dialog for selecting trials
	CTrialSelectionDialog::Selection available;
	Tokenizer tok;                               //ASC text file tokenizer
	std::list<void*> tempList;                   //temp list of trials
	std::ostringstream errStrm;                  //error message stream
	std::string message;
	TrialData *currentTrial;
	short index = 0;

	tok.open(dataFileName);
	if(tok.isOkay() == false) {                  //failed to open file, throw
		errStrm<<"Unable to open file '"<<dataFileName<<"' for import."<<std::endl;
		throw(std::exception(errStrm.str().c_str()));
	}
	else {                                       //init error stream and progress bar
		errStrm<<"Error importing file '"<<dataFileName<<"':"<<std::endl;

		progressDialog.Create(IDD_PROGRESS_DIALOG, 0);
		progressDialog.SetWindowText(("Importing " + dataFileName).c_str());
		progressDialog.SetRange(0, tok.fileLength());
		progressDialog.SetPosition(0);
	}

	while(tok.isOkay() && _displaySize.IsRectEmpty()) {
		Tokenizer::EventType event = tok.readEventType();

		if(event == Tokenizer::MESSAGE) {        //look for the DISPLAY COORDS message
                                                 //update progress bar
			progressDialog.SetPosition(tok.filePosition());

			tok.readUnsignedInt();               //skip the time stamp, read display size
			if(tok.readMessageType() == Tokenizer::DISPLAY_SIZE) {
				_displaySize.left = tok.readShortInt();
				_displaySize.top = tok.readShortInt();
				_displaySize.right = tok.readShortInt();
				_displaySize.bottom = tok.readShortInt();
			}
		}
	}

	if(!tok.isOkay()) {
		errStrm<<"Unable to find or read DISPLAY COORDS data."<<std::endl;
		throw(std::exception(errStrm.str().c_str()));
	}

	while(tok.isOkay()) {                        //read the trials
		Tokenizer::EventType event = tok.readEventType();

		if(event == Tokenizer::MESSAGE) {
			Tokenizer::MessageType msg;
			
			tok.readUnsignedInt();
			msg = tok.readMessageType();         //skip the time stamp
			                                     //update progress bar
			progressDialog.SetPosition(tok.filePosition());

			if(msg == Tokenizer::TRIAL_ID) {     //read trial
				std::string trialId = tok.readString();
				std::string message("Importing trial ");
				if(trialId == "E1I5D0"){
					message.append("error");
				}

				message.append(trialId);         //update progress bar message
				progressDialog.SetMessage(message.c_str());

				currentTrial = new TrialData(trialId);
				currentTrial->import(tok, _jhookThreshold);
				tempList.push_back(currentTrial);
				currentTrial->setDisplayCoords(_displaySize);

				const TrialId &id = currentTrial->trialID();
				if(id.condition() > available.condMax)
					available.condMax = id.condition();
				if(id.item() > available.itemMax)
					available.itemMax = id.item();
				if(id.index() > available.depMax)
					available.depMax = id.index();
				if(id.isExperimental())
					available.useExperimental = true;
				if(id.isFiller())
					available.useFiller = true;
				if(id.isPractice())
					available.usePractice = true;
			}

			else if(msg == Tokenizer::TRIAL_STATUS) { //read status
				std::string status = tok.readString();
				if(status.compare("ABORTED") == 0) 
					currentTrial->setFlag(TRIAL_ABORTED, true);
				else if(status.compare("ERROR") == 0)
					currentTrial->setFlag(TRIAL_ERROR, true);
				else if(status.compare("REPEATED") == 0)
					currentTrial->setFlag(TRIAL_REPEATED, true);
			}
		}
		else if (event == Tokenizer::TRIAL_END && (currentTrial->testFlag(TRIAL_NO_END_MESSAGE))){
			currentTrial->setFlag(TRIAL_ERROR, false);
			// then do nothing, because we missed this
			// message when processing the previous trial 
			// due to message asynchronicity.
		}
	}

	dialog.SetSelections(available);
	if(dialog.DoModal() == IDOK)                 //let user select which trials to edit
		available = dialog.GetSelections();

	_trial.reserve(tempList.size());             //convert list of trials to vector
	while(!tempList.empty()) {
		TrialData *d = static_cast<TrialData*>(tempList.front());
		const TrialId &id = d->trialID();
		bool selected = true;

		if(id.condition() < available.condMin || id.condition() > available.condMax)
			selected = false;
		if(id.item() < available.itemMin || id.item() > available.itemMax)
			selected = false;
		if(id.index() < available.depMin || id.index() > available.depMax)
			selected = false;
		if(id.isExperimental() && available.useExperimental == false)
			selected = false;
		if(id.isFiller() && available.useFiller == false)
			selected= false;
		if(id.isPractice() && available.usePractice == false)
			selected = false;

		if(selected) {
			_trial.push_back(d);
			d->setIndex(++index);
		}
		else delete d;
		tempList.pop_front();
	}

	if(index == 0) {
		errStrm<<"No trials selected!"<<std::endl;
		throw(std::exception(errStrm.str().c_str()));
	}
}

void ExperimentData::import(const std::string &dataFileName, const std::string &senFileName)
	throw(std::exception)
{
	CProgressDialog progressDialog;              //dialog for the progress bar
	CTrialSelectionDialog dialog;
	CTrialSelectionDialog::Selection available;
	Tokenizer dataTok, senTok;
	std::ostringstream os;
	std::list<void*> tempList;
	int firstLine = 0;

	dataTok.open(dataFileName);                  //open data file for reading
	if(dataTok.isOkay() == false) {
		os<<"Unable to open file '"<<dataFileName<<"' for import."<<std::endl;
		throw(std::exception(os.str().c_str()));
	}

	senTok.open(senFileName);                    //open sentence file for reading
	if(senTok.isOkay() == false) {
		os<<"Unable to open sentence file '"<<senFileName<<"' for import."<<std::endl;
		throw(std::exception(os.str().c_str()));
	}

	os<<"Error importing sentence file '"<<senFileName<<"':"<<std::endl;

	progressDialog.Create(IDD_PROGRESS_DIALOG, 0);
	progressDialog.SetWindowText(("Importing " + senFileName).c_str());
	progressDialog.SetRange(0, senTok.fileLength());
	progressDialog.SetPosition(0);

	firstLine = senTok.readInteger();
	while(senTok.isOkay()) {                     //read the sentence file
		TrialData *td;
		std::ostringstream trialId;
		int cond, item;
		char ch;

		if((ch = senTok.readChar()) != '{') {
			os<<"Expected '{' before condition and item numbers, found "<<ch;
			os<<" instead on line "<<senTok.lineNumber()<<std::endl;
			throw(std::exception(os.str().c_str()));
		}

		cond = senTok.readInteger();             //read the trial identifier (cond & item)
		item = senTok.readInteger();
		trialId<<'C'<<cond<<'I'<<item;

		if((ch = senTok.readChar()) != '}' || !senTok.isOkay()) {
			os<<"Expected '}' after condition and item numbers, found "<<ch;
			os<<" instead on line "<<senTok.lineNumber()<<std::endl;
			throw(std::exception(os.str().c_str()));
		}

		progressDialog.SetMessage(("Importing SEN for " + trialId.str()).c_str());
		progressDialog.SetPosition(senTok.filePosition());

		td = new TrialData(trialId.str());       //initialize a trial
		td->importSen(senTok);                   //read a single sentence (trial)
		tempList.push_back(td);

		if(cond > available.condMax)
			available.condMax =cond;
		if(item > available.itemMax)
			available.itemMax = item;
	}

	available.useExperimental = true;
	dialog.SetSelections(available);
	if(dialog.DoModal() == IDOK)                 //let user select which trials to edit
		available = dialog.GetSelections();

	os.str("");
	os<<"Error importing DPI data file '"<<dataFileName<<"':"<<std::endl;
	_trial.reserve(tempList.size());             //convert list of trials to vector

	progressDialog.SetWindowText(("Importing " + dataFileName).c_str());
	progressDialog.SetRange(0, dataTok.fileLength());
	progressDialog.SetPosition(0);

	while(dataTok.isOkay() && !tempList.empty()) {
		TrialData *trial = static_cast<TrialData*>(tempList.front());
		std::ostringstream trialId;
		int cond, index, item;

		index = dataTok.readInteger();           //read trial index and identifier
		cond = dataTok.readInteger();
		item = dataTok.readInteger();

		trialId<<'C'<<cond<<'I'<<item;

		if(trialId.str().compare(trial->trialID().idString()) == 0) {
			bool selected = true;

			if(cond < available.condMin || cond > available.condMax)
				selected = false;
			if(item < available.itemMin || item > available.itemMax)
				selected = false;

			if(selected) {
				progressDialog.SetMessage(("Importing DAT for " + trialId.str()).c_str());
				progressDialog.SetPosition(dataTok.filePosition());

				trial->importData(dataTok, firstLine);
				trial->setIndex(index);
				_trial.push_back(trial);
			}
			else {
				dataTok.readLine();
				delete trial;
			}
			tempList.pop_front();
		}
		else dataTok.readLine();
	}

	if(!tempList.empty()) {
		os<<"Reached end of data file without acquiring data for all trials."<<std::endl;
		throw(std::exception(os.str().c_str()));
	}
}

int ExperimentData::nTrials() const
{
	return(static_cast<int>(_trial.size()));
}

void ExperimentData::read(const std::string &fileName) throw(std::exception)
{
	CProgressDialog progressDialog;              //dialog for the progress bar
	std::ifstream ifs(fileName.c_str(), std::ios::in | std::ios::binary);
	std::string version;
	int len, nTrials, vLen = -1, vNum = 0;

	if(!ifs.is_open())
		throw(std::exception("Error opening file for input\n"));

	ifs.seekg(0, std::ios::end);                 //compute length of file
	len = static_cast<int>(ifs.tellg());
	ifs.seekg(0, std::ios::beg);

	getline(ifs, version, '!');                 //read version info
	if(version.compare("EDD_0.2.0") == 0) vNum = 2;
	else if(version.compare("EDD_0.3.0") == 0) vNum = 3;
	else if(version.compare("EDD_0.4.0") == 0) vNum = 4;
	else if(version.compare("EDD_0.5.0") == 0) vNum = 5;
	else throw(std::exception("Error reading input file: Missing or bad EDD version number\n"));
	                                             //read display size
	ifs.read(reinterpret_cast<char*>(&_displaySize), sizeof(CRect));
	ifs.read(reinterpret_cast<char*>(&nTrials), sizeof(int));

	progressDialog.Create(IDD_PROGRESS_DIALOG, 0);
	progressDialog.SetWindowText(("Reading " + fileName).c_str());
	progressDialog.SetRange(0, len);
	progressDialog.SetPosition(0);

	_trial.reserve(nTrials);                     //allocate vector of trials
	for(int i = 0; i < nTrials; ++i) {
		std::string message("Read trial ");
		TrialData *trial = new TrialData;

		trial->read(ifs, vNum);
		if(vNum == 2)
			trial->setIndex(i);
        _trial.push_back(trial);

		message.append(trial->trialID().idString());
		progressDialog.SetMessage(message.c_str());
		progressDialog.SetPosition(static_cast<int>(ifs.tellg()));
	}

	ifs.close();
}

void ExperimentData::revert()
{
	for(unsigned int i = 0; i < _trial.size(); ++i)
		_trial[i]->revert();
}

TrialData *ExperimentData::trial(unsigned int index)
{
	if(index < _trial.size())
		return(_trial[index]);
	return(0);
}

void ExperimentData::write(const std::string &fileName) const throw(std::exception)
{
	std::ofstream ofs(fileName.c_str(), std::ios::out | std::ios::binary);
	int nTrials = static_cast<int>(_trial.size());

	if(!ofs.is_open())
		throw(std::exception("Error opening file for output\n"));

	ofs.write("EDD_0.5.0!", 10);                 //write version number
                                                 //write display size
	ofs.write(reinterpret_cast<const char*>(&_displaySize), sizeof(CRect));
	ofs.write(reinterpret_cast<const char*>(&nTrials), sizeof(int));

	for(unsigned int i = 0; i < _trial.size(); ++i) //write trials
		_trial[i]->write(ofs);

	ofs.close();
}

void ExperimentData::writeSummary(std::ostream &os, bool tabbed) const
{
	float nCorrect = 0, nQstn = 0;

	for(unsigned int i = 0; i < _trial.size(); ++i) {
		short status = _trial[i]->questionStatus();

		if(status > 0)
			++nCorrect;
		if(status != 0)
			++nQstn;
		_trial[i]->writeSummary(os, tabbed);
	}

	if(!tabbed) {
		os<<"Total correct subject responses: "<<nCorrect<<"\r\n";
		os<<"Total questions:                 "<<nQstn<<"\r\n";
		os<<"Subject Accuracy:                "<<(nCorrect / nQstn)<<"\r\n";
	}
}

void ExperimentData::set_jhook_threshold(int jh){
	_jhookThreshold = jh;
}
