/*  Eyetracking software creates and executes experiments using 
    Eyelink II hardware and libraries by SR Research

    Copyright (C) 2004  David J. Stracuzzi, University of Massachusetts

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "StdAfx.h"
#include "EyeEvent.h"

EyeEvent::EyeData::EyeData()
  : startTime(0), endTime(0), 
    isBoundary(false), isDeleted(false), isRight(false), isTrackLoss(false),
	crossesDCboundary(false)
{}

const EyeEvent::EyeData &EyeEvent::EyeData::operator=(const EyeEvent::EyeData &rhs)
{
	startTime   = rhs.startTime;
	endTime     = rhs.endTime;
	isBoundary  = rhs.isBoundary;
	isDeleted   = rhs.isDeleted;
	isRight     = rhs.isRight;
	isTrackLoss = rhs.isTrackLoss;
	crossesDCboundary = rhs.crossesDCboundary;
	return(*this);
}

/*******************************************************************************************/

EyeEvent::EyeEvent()
  : _ed(), _edBackup(), _next(0), _prev(0)
    
{}

EyeEvent::EyeEvent(bool rightEye, bool trackLoss, bool boundary, unsigned int start, unsigned int end)
  : _ed(), _edBackup(), _next(0), _prev(0)
{
	/*_ed.startTime   = _edBackup.startTime   = start;
	_ed.endTime     = _edBackup.endTime     = end;
	_ed.isBoundary  = _edBackup.isBoundary  = boundary;
	_ed.isDeleted   = _edBackup.isDeleted   = false;
	_ed.isRight     = _edBackup.isRight     = rightEye;
	_ed.isTrackLoss = _edBackup.isTrackLoss = trackLoss;*/

	_ed.startTime   = start;
	_ed.endTime     = end;
	_ed.isBoundary  = boundary;
	_ed.isDeleted   =  false;
	_ed.isRight     = rightEye;
	_ed.isTrackLoss =  trackLoss;

	_edBackup.startTime   = start;
	_edBackup.endTime     = end;
	_edBackup.isBoundary  = boundary;
	_edBackup.isDeleted   = false;
	_edBackup.isRight     = rightEye;
	_edBackup.isTrackLoss = trackLoss;


}

EyeEvent::EyeEvent(const EyeEvent &event)
{
	_ed       = EyeData(event._ed);
	_edBackup = EyeData(event._edBackup);
	_next     = event._next;
	_prev     = event._prev;
}

EyeEvent::~EyeEvent()
{}

void EyeEvent::deleteEvent()
{
	_ed.isDeleted = true;
}

void EyeEvent::draw(CDC *pDC, const Colors &colors, const StateFlags &flags)
{}

unsigned int EyeEvent::duration() const
{
	return(_ed.endTime - _ed.startTime);
}

unsigned int EyeEvent::endTime() const
{
	return(_ed.endTime);
}

void EyeEvent::export(std::ostream &os) const
{}

bool EyeEvent::isDeleted() const
{
	return(_ed.isDeleted);
}

bool EyeEvent::isRightEye() const
{
	return(_ed.isRight);
}

bool EyeEvent::isTrackLoss() const
{
	return(_ed.isTrackLoss);
}

int EyeEvent::length() const
{
	if(_next) return(_next->length() + 1);
	else return(1);
}

EyeEvent *EyeEvent::next(bool ignoreDeleted) const
{
	if(ignoreDeleted) {
		EyeEvent *ePtr = _next;
		while(ePtr && ePtr->_ed.isDeleted)
			ePtr = ePtr->_next;
		return(ePtr);
	}
	else return(_next);
}

const EyeEvent& EyeEvent::operator=(const EyeEvent &rhs)
{
	_ed       = rhs._ed;
	_edBackup = rhs._edBackup;
	_next     = rhs._next;
	_prev     = rhs._prev;
	return(*this);
}

EyeEvent *EyeEvent::previous(bool ignoreDeleted) const
{
	if(ignoreDeleted) {
		EyeEvent *ePtr = _prev;
		while(ePtr && ePtr->_ed.isDeleted)
			ePtr = ePtr->_prev;
		return(ePtr);
	}
	else return(_prev);
}

void EyeEvent::read(std::istream &is, int versionNum)
{                                                //read event and backup data
	static const int sz = sizeof(EyeEvent::EyeData);

	is.read(reinterpret_cast<char*>(&_ed), sz);
	is.read(reinterpret_cast<char*>(&_edBackup), sz);

	if(versionNum < 3) {
		_ed.isTrackLoss = _ed.isRight;
		_ed.isRight = _ed.isDeleted;
		_ed.isDeleted = _ed.isBoundary;
		_ed.isBoundary = false;

		_edBackup.isTrackLoss = _edBackup.isRight;
		_edBackup.isRight = _edBackup.isDeleted;
		_edBackup.isDeleted = _edBackup.isBoundary;
		_edBackup.isBoundary = false;
	}
}

void EyeEvent::revert()
{
	_ed = _edBackup;
}

void EyeEvent::setNext(EyeEvent *event)
{
	_next = event;
	event->_prev = this;
}

void EyeEvent::setPrevious(EyeEvent *event)
{
	_prev = event;
	event->_next = this;
}

unsigned int EyeEvent::startTime() const
{
	return(_ed.startTime);
}

void EyeEvent::truncateFutureEvents()
{
	EyeEvent *ePtr = _next;
	while(ePtr) {
		ePtr->_ed.isDeleted = true;
		ePtr = ePtr->_next;	
	}
}

void EyeEvent::updateInitial(int deltaStartTime, int deltaEndTime) throw(std::exception)
{
	if(_ed.endTime + deltaEndTime < _ed.startTime + deltaStartTime)
		throw(std::exception("Event cannot have negative duration.\nTime adjustment action ignored."));

	_ed.endTime += deltaEndTime;
	_ed.startTime += deltaStartTime;
	_edBackup.startTime += deltaStartTime;
	_edBackup.endTime += deltaEndTime;
}


void EyeEvent::update(int deltaStartTime, int deltaEndTime) throw(std::exception)
{
	if(_ed.endTime + deltaEndTime < _ed.startTime + deltaStartTime)
		throw(std::exception("Event cannot have negative duration.\nTime adjustment action ignored."));

	_ed.endTime += deltaEndTime;
	_ed.startTime += deltaStartTime;
}

void EyeEvent::write(std::ostream &os) const
{
	static const int sz = sizeof(EyeEvent::EyeData);
                                                 //write event and backup data
	os.write(reinterpret_cast<const char*>(&_ed), sz);
	os.write(reinterpret_cast<const char*>(&_edBackup), sz);
}
