/*  Eyetracking software creates and executes experiments using 
    Eyelink II hardware and libraries by SR Research

    Copyright (C) 2004  David J. Stracuzzi, University of Massachusetts

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "StdAfx.h"
#include "FixationEvent.h"

FixationEvent::FixData::FixData()
  : charLoc(-1, -1), pixelLoc(0, 0), rectLoc(0, 0, 0, 0), index(0)
{}

const FixationEvent::FixData &
FixationEvent::FixData::operator=(const FixationEvent::FixData &rhs)
{
	charLoc  = rhs.charLoc;
	pixelLoc = rhs.pixelLoc;
	rectLoc  = rhs.rectLoc;
	index    = rhs.index;
	return(*this);
}

/*****************************************************************************************/

short FixationEvent::_leftIndex = 0;
short FixationEvent::_rightIndex = 0;

FixationEvent::FixationEvent()
  : EyeEvent(), _fd(), _fdBackup()
{}

FixationEvent::FixationEvent(bool rightEye, bool trackLoss, bool boundary,
							 unsigned int start, unsigned int end, 
							 short x, short y)
  : EyeEvent(rightEye, trackLoss, boundary, start, end), _fd(), _fdBackup()
{
	_fd.pixelLoc.SetPoint(x, y);
	if(_ed.isRight) _fd.index = ++_rightIndex;
	else _fd.index = ++_leftIndex;

	_fdBackup = _fd;
}

FixationEvent::FixationEvent(const FixationEvent &event)
  : EyeEvent(event)
{
	_fd       = event._fd;
	_fdBackup = event._fdBackup;
}

FixationEvent::~FixationEvent()
{}

const CPoint &FixationEvent::charLocation() const
{
	return(_fd.charLoc);
}

void FixationEvent::draw(CDC *pDC, const Colors &colors, const StateFlags &flags)
{
	if(!flags.test(EVENT_FIXATION))              //do not draw fixations
		_fd.rectLoc.SetRectEmpty();
	else {
		COLORREF localColor;
		CSize textSz;
		char buf[10];
		int length = 0;
		bool badFix = (duration() < 80 || duration() > 800);

		_itoa(_fd.index, buf, 10);               //convert index to a string
		while(buf[length] != '\0') ++length;

		textSz = pDC->GetTextExtent(buf, length); //compute enclosing rectangle
		_fd.rectLoc.SetRect(_fd.pixelLoc.x, _fd.pixelLoc.y,
			             _fd.pixelLoc.x + textSz.cx, _fd.pixelLoc.y + textSz.cy);
		_fd.rectLoc.OffsetRect(-textSz.cx/2, -textSz.cy/2);

		if(_ed.isDeleted) {                      //this is a deleted event
			if(flags.test(HIGHLIGHT_DELETED)) {  //highlight it
				COLORREF color = _ed.isRight ? colors.rightEyeHlt : colors.leftEyeHlt;
				CPen *pOldPen, pen(PS_SOLID, 2, color);

				pOldPen = pDC->SelectObject(&pen);
				pDC->MoveTo(_fd.rectLoc.left, _fd.rectLoc.top);
				pDC->LineTo(_fd.rectLoc.right, _fd.rectLoc.top);
				pDC->LineTo(_fd.rectLoc.right, _fd.rectLoc.bottom);
				pDC->LineTo(_fd.rectLoc.left, _fd.rectLoc.bottom);
				pDC->LineTo(_fd.rectLoc.left, _fd.rectLoc.top);
				pDC->LineTo(_fd.rectLoc.right, _fd.rectLoc.bottom);
				pDC->MoveTo(_fd.rectLoc.right, _fd.rectLoc.top);
				pDC->LineTo(_fd.rectLoc.left, _fd.rectLoc.bottom);
				pDC->SelectObject(pOldPen);
			}
			else _fd.rectLoc.SetRectEmpty();     //do not highlight it
			return;
		}

		if(flags.test(SELECTED)) {               //this event is selected, draw saccades
			SaccadeEvent *lead = leadSaccade();
			SaccadeEvent *trail = trailSaccade();
			
			localColor = colors.selected;
			if(lead) lead->draw(pDC, colors, flags);
			if(trail) trail->draw(pDC, colors, flags);
		}
		else if(flags.test(HIGHLIGHT_BLINK) && _ed.isTrackLoss) //this event has a blink
			localColor = _ed.isRight ? colors.rightEye : colors.leftEye;
		else if(_ed.isBoundary)                  //this crosses the boundary on preview/fast prime
			localColor = colors.postDisplayChange;         //  (data is always right eye only
		else if(flags.test(HIGHLIGHT_FIXATION) && badFix)       //this event too long/short
			localColor = _ed.isRight ? colors.rightEyeHlt : colors.leftEyeHlt;
		else localColor = _ed.isRight ? colors.rightEye : colors.leftEye;
	
		if(flags.test(HIGHLIGHT_BLINK) && _ed.isTrackLoss) {    //draw fixation as a circle
			CBrush *pOldBrush, brush(localColor);
			CPen *pOldPen, pen(PS_SOLID, 1, localColor);
			CRect r(_fd.rectLoc);

			r.InflateRect(r.Height()/2, r.Width()/2, r.Height()/2, r.Width()/2);
			pOldBrush = pDC->SelectObject(&brush);
			pOldPen = pDC->SelectObject(&pen);
			pDC->Ellipse(&r);
			pDC->SelectObject(pOldBrush);
			pDC->SelectObject(pOldPen);
		}

		pDC->SetBkMode(OPAQUE);
		pDC->SetBkColor(localColor);
		pDC->SetTextColor(colors.label);           //draw text with opaque background
		pDC->DrawText(buf, length, &_fd.rectLoc, DT_LEFT | DT_TOP | DT_NOCLIP | DT_NOPREFIX);
	}
}

void FixationEvent::export(std::ostream &os) const
{
	if(_ed.isDeleted == true)
		os<<' '<<_fd.charLoc.x<<" 99 "<<_ed.startTime<<' '<<_ed.endTime;
	else
		os<<' '<<_fd.charLoc.x<<' '<<_fd.charLoc.y<<' '<<_ed.startTime<<' '<<_ed.endTime;
}

short FixationEvent::index() const
{
	return(_fd.index);
}

void FixationEvent::initialize(bool initRight)
{
	if(initRight)
		_rightIndex = 0;
	else 
		_leftIndex = 0;
}

SaccadeEvent *FixationEvent::leadSaccade() const
{
	EyeEvent *ePtr = previous();

	while(ePtr) {
		SaccadeEvent *event = dynamic_cast<SaccadeEvent*>(ePtr);
		if(event) return(event);
		else ePtr = ePtr->previous();
	}
	return(0);
}

void FixationEvent::mergeWithFixation(bool mergeWithPrev)
{
	EyeEvent *ePtr = mergeWithPrev ? previous() : next();

	while(ePtr) {
		FixationEvent *event = dynamic_cast<FixationEvent*>(ePtr);
		if(event) {                              //found previous fixation
			if(duration() < event->duration()) { //use location of longer fixation
				_fd.charLoc = event->_fd.charLoc;
				_fd.pixelLoc = event->_fd.pixelLoc;
				_fd.rectLoc = event->_fd.rectLoc;
			}
			if(mergeWithPrev)                    //merged event covers entire time span
				_ed.startTime = event->_ed.startTime;
			else _ed.endTime = event->_ed.endTime;
			event->_ed.isDeleted = true;         //delete the previous
			break;
		}
		else ePtr->deleteEvent();                //delete intervening saccades
		ePtr = mergeWithPrev ? ePtr->previous() : ePtr->next();
	}
}

void FixationEvent::mergeWithSaccade()
{
	EyeEvent *ePtr = next();

	_ed.isDeleted = true;
	if(dynamic_cast<SaccadeEvent*>(ePtr)) {      //if preceeding event is a saccade
		int diff = ePtr->startTime() - _ed.startTime; //update start time
		ePtr->update(-diff, 0);
	}
}

const FixationEvent &FixationEvent::operator=(const FixationEvent &rhs)
{
	_ed       = rhs._ed;
	_edBackup = rhs._edBackup;
	_next     = rhs._next;
	_prev     = rhs._prev;
	_fd       = rhs._fd;
	_fdBackup = rhs._fdBackup;
	return(*this);
}

const CPoint &FixationEvent::pixelLocation() const
{
	return(_fd.pixelLoc);
}

void FixationEvent::read(std::istream &is, int versionNum)
{
	static const int sz = sizeof(FixationEvent::FixData);
		
	EyeEvent::read(is, versionNum);              //read basic event data

	is.read(reinterpret_cast<char*>(&_fd), sz);  //read fixation and backup data
	is.read(reinterpret_cast<char*>(&_fdBackup), sz);
}

CRect FixationEvent::rectLocation() const
{
	return(_fd.rectLoc);
}

void FixationEvent::revert()
{
	EyeEvent::revert();
	_fd = _fdBackup;
}

SaccadeEvent *FixationEvent::trailSaccade() const
{
	EyeEvent *ePtr = next();

	while(ePtr) {
		SaccadeEvent *event = dynamic_cast<SaccadeEvent*>(ePtr);
		if(event) return(event);
		else ePtr = ePtr->next();
	}
	return(0);
}

void FixationEvent::testBlink()
{
	if(_ed.endTime - _ed.startTime < 120) {
		if(_next && _next->isTrackLoss())
			_ed.isTrackLoss = _edBackup.isTrackLoss = true;
		if(_prev && _prev->isTrackLoss())
			_ed.isTrackLoss = _edBackup.isTrackLoss = true;
	}
}

void FixationEvent::update(const CPoint &rgnLoc)
{
	_fd.charLoc = rgnLoc;
}

void FixationEvent::updateInitial(const CPoint &rgnLoc)
{
	_fd.charLoc = rgnLoc;
	_fdBackup.charLoc = rgnLoc;
}


void FixationEvent::update(int deltaX, int deltaY)
{
	SaccadeEvent *lead = leadSaccade();
	SaccadeEvent *trail = trailSaccade();

	_fd.pixelLoc.x += deltaX;
	_fd.pixelLoc.y += deltaY;

	if(lead) lead->update(deltaX, deltaY, false);
	if(trail) trail->update(deltaX, deltaY, true);
}

void FixationEvent::write(std::ostream &os) const
{
	static const int sz = sizeof(FixationEvent::FixData);

	os.put('F');                                 //mark this as a fixation
	EyeEvent::write(os);                         //read basic event data
                                                 //write fixation and backup data
	os.write(reinterpret_cast<const char*>(&_fd), sz);
	os.write(reinterpret_cast<const char*>(&_fdBackup), sz);
}
