/*  Eyetracking software creates and executes experiments using 
    Eyelink II hardware and libraries by SR Research

    Copyright (C) 2004  David J. Stracuzzi, University of Massachusetts

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "StdAfx.h"
#include "SaccadeEvent.h"

SaccadeEvent::SacData::SacData()
  : startPixel(0, 0), endPixel(0, 0)
{}

const SaccadeEvent::SacData &
SaccadeEvent::SacData::operator=(const SaccadeEvent::SacData &rhs)
{
	startPixel = rhs.startPixel;
	endPixel   = rhs.endPixel;
	return(*this);
}

/**************************************************************************************/

SaccadeEvent::SaccadeEvent()
  : EyeEvent(), _sd(), _sdBackup()
{}

SaccadeEvent::SaccadeEvent(bool rightEye, bool trackLoss, bool boundary,
						   unsigned int startTime, unsigned int endTime,
						   short x1, short y1, short x2, short y2)
  : EyeEvent(rightEye, trackLoss, boundary, startTime, endTime), _sd(), _sdBackup()
{
	_sd.startPixel.SetPoint(x1, y1);
	_sd.endPixel.SetPoint(x2, y2);
	if(x1 == -1 && x2 == -1 && y1 == -1 && y2 == -1)
		_ed.isDeleted = _edBackup.isDeleted = true;

	_sdBackup = _sd;
}

SaccadeEvent::SaccadeEvent(const SaccadeEvent &event)
  : EyeEvent(event)
{
	_sd       = event._sd;
	_sdBackup = event._sdBackup;
}

SaccadeEvent::~SaccadeEvent(void)
{}

void SaccadeEvent::draw(CDC *pDC, const Colors &colors, const StateFlags &flags)
{
	if(flags.test(EVENT_SACCADE) && !_ed.isDeleted) {
		COLORREF localColor = _ed.isRight ? colors.rightEye : colors.leftEye;
		CPen pen, *pOldPen;
		bool badSacc = (/*duration() < 10 ||*/ duration() > 50);

		if(flags.test(SELECTED))                 //this event is selected
			pen.CreatePen(PS_SOLID, 2, colors.selected);
		else if(_ed.isBoundary)                  //this event crosses boundary in preview/fast prime
			localColor = colors.leftEye;
		else if(flags.test(HIGHLIGHT_BLINK) && _ed.isTrackLoss) { //this event has a blink
			localColor = _ed.isRight ? colors.rightEye : colors.leftEye;
			pen.CreatePen(PS_SOLID, 5, localColor);
		}
		else if(flags.test(HIGHLIGHT_SACCADE) && badSacc) { //this event is to long/short
			localColor = _ed.isRight ? colors.rightEyeHlt : colors.leftEyeHlt;
			pen.CreatePen(PS_SOLID, 5, localColor);
		}
		else pen.CreatePen(PS_SOLID, 1, localColor);

		pOldPen = pDC->SelectObject(&pen);
		pDC->MoveTo(_sd.startPixel);
		pDC->LineTo(_sd.endPixel);
		pDC->SelectObject(pOldPen);
	}
}

CPoint SaccadeEvent::endPixel() const
{
	return(_sd.endPixel);
}

void SaccadeEvent::export(std::ostream &os) const
{}

const SaccadeEvent &SaccadeEvent::operator=(const SaccadeEvent &rhs)
{
	_ed       = rhs._ed;
	_edBackup = rhs._edBackup;
	_next     = rhs._next;
	_prev     = rhs._prev;
	_sd       = rhs._sd;
	_sdBackup = rhs._sdBackup;
	return(*this);
}

void SaccadeEvent::read(std::istream &is, int versionNum)
{
	static const int sz = sizeof(SaccadeEvent::SacData);

	EyeEvent::read(is, versionNum);              //read basic event data

	is.read(reinterpret_cast<char*>(&_sd), sz);  //read saccade and backup data
	is.read(reinterpret_cast<char*>(&_sdBackup), sz);
}

void SaccadeEvent::revert()
{
	EyeEvent::revert();
	_sd = _sdBackup;
}

CPoint SaccadeEvent::startPixel() const
{
	return(_sd.startPixel);
}

void SaccadeEvent::update(int deltaX, int deltaY, bool applyToStartPixel)
{
	if(applyToStartPixel) {
		_sd.startPixel.x += deltaX;
		_sd.startPixel.y += deltaY;
	}
	else {
		_sd.endPixel.x += deltaX;
		_sd.endPixel.y += deltaY;
	}
}

void SaccadeEvent::write(std::ostream &os) const
{
	static const int sz = sizeof(SaccadeEvent::SacData);

	os.put('S');                                 //mark this as a saccade
	EyeEvent::write(os);                         //write basic event data
                                                 //write saccade and backup data
	os.write(reinterpret_cast<const char*>(&_sd), sz);
	os.write(reinterpret_cast<const char*>(&_sdBackup), sz);
}
