/*  Eyetracking software creates and executes experiments using 
    Eyelink II hardware and libraries by SR Research

    Copyright (C) 2004  David J. Stracuzzi, University of Massachusetts

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "StdAfx.h"
#include <algorithm>
#include <sstream>
#include "ExperimentData.h"
#include "SelectionState.h"
#include "TrialData.h"

SelectionState::SelectionState()
  : _experiment(0), _flags(), _labelFontHt(8), _trialPtr(0), _jhookThreshold(10)
{
	_colors.background = 0xFFFFFF;
	_colors.label = 0xFFFFFF;
	_colors.leftEye = 0xAA0000;
	_colors.leftEyeHlt = 0xDD0000;
	_colors.region = 0xAFAFAF;
	_colors.rightEye = 0x009900;
	_colors.rightEyeHlt = 0x00FF00;
	_colors.selected = 0x0000FF;
	_colors.stimulus = 0x000000;
	_colors.postDisplayChange = 0xCCAA00;

	_flags.set();                                //sets all flags to true
	_flags.set(SELECTED, false);                 //unset the SELECTED flag initially
}

SelectionState::~SelectionState()
{
	if(_experiment) delete _experiment;
}

void SelectionState::clear()
{
	if(_experiment) {
		delete _experiment;
		_experiment = 0;
	}
	_selected.clear();
	_filename.clear();
	_trialPtr = 0;
}

Colors &SelectionState::colors()
{
	return(_colors);
}

FixationEvent *SelectionState::currFixation()
{                                                //return most recent active selection (if any)
	if(_experiment && _flags.test(EVENT_FIXATION) && !_selected.empty())
		return(_selected.back());
	return(0);
}

const ExperimentData *SelectionState::experiment() const
{
	return(_experiment);
}

std::string SelectionState::exportExperiment(const std::string &filename, 
											 const ExportParams &params) const throw(std::exception)
{
	std::ostringstream os;

	_experiment->export(filename, os, params);
	return(os.str());
}

const std::string &SelectionState::filename() const
{
	return(_filename);
}

const StateFlags &SelectionState::flags() const
{
	return(_flags);
}

bool SelectionState::hasExperiment() const
{
	return(_experiment != 0);
}

void SelectionState::importExperiment(const std::string &filename)
{
	std::string::size_type pos;

	clear();
	_experiment = new ExperimentData();
	_experiment->set_jhook_threshold(_jhookThreshold);
	_experiment->import(filename);

	_filename = filename;
	pos = filename.find(".ASC");
	if(pos == std::string::npos)
		pos = filename.find(".asc");
	if(pos == std::string::npos)
		_filename.clear();
	else _filename.erase(pos);
}

void SelectionState::importExperiment(const std::string &datafile, const std::string &senfile)
{
	std::string::size_type pos;

	clear();
	_experiment = new ExperimentData();
	_experiment->import(datafile, senfile);

	pos = senfile.find(".SEN");
	_filename = senfile;
	if(pos == std::string::npos)
		pos = senfile.find(".sen");
	if(pos == std::string::npos)
		_filename.clear();
	else _filename.erase(pos);
}

int SelectionState::labelFontHt() const
{
	return(_labelFontHt);
}

int SelectionState::jhookThreshold() const
{
	return(_jhookThreshold);
}

bool SelectionState::useMultiChange() const
{
	return(_useMultiChange);
}

void SelectionState::moveTrialPtr(short delta)
{
	const maxTrials = _experiment->nTrials();

	_trialPtr += delta;
	if(_trialPtr < 0)
		_trialPtr = 0;
	else if(_trialPtr >= maxTrials)
		_trialPtr = maxTrials-1;

	_selected.clear();
}

FixationEvent *SelectionState::nextFixation(bool multiSelect)
{
	if(_experiment && _flags.test(EVENT_FIXATION) && !_selected.empty())  {
		EyeEvent *ePtr = _selected.back();       //use most recent active selection

		if(!multiSelect)                         //clear list of selected fixations
			_selected.clear();

		ePtr = ePtr->next();
		while(ePtr) {                            //find next fixation
			FixationEvent *event = dynamic_cast<FixationEvent*>(ePtr);

			if(event) {                          //if event == 0 then it was not a Fixation
				if(std::find(_selected.begin(), _selected.end(), ePtr) != _selected.end())
					_selected.pop_back();
				else _selected.push_back(event);
				return(event);
			}
			else ePtr = ePtr->next();
		}
	}
	return(0);
}

void SelectionState::openExperiment(const std::string &filename) throw(std::exception)
{
	std::string::size_type pos;

	clear();

	_experiment = new ExperimentData();
	_experiment->read(filename);

	_filename = filename;
	pos = filename.find(".EDD");
	if(pos == std::string::npos)
		pos = filename.find(".edd");
	if(pos != std::string::npos)
		_filename.erase(pos);
}

FixationEvent *SelectionState::prevFixation(bool multiSelect)
{
	if(_experiment && _flags.test(EVENT_FIXATION) && !_selected.empty()) {
		EyeEvent *ePtr = _selected.back();       //use most recent active selection

		if(!multiSelect)                         //clear list of selected fixations
			_selected.clear();

		ePtr = ePtr->previous();
		while(ePtr) {                            //find next fixation
			FixationEvent *event = dynamic_cast<FixationEvent*>(ePtr);

			if(event) {                          //if event == 0 then it was not a Fixation
				if(std::find(_selected.begin(), _selected.end(), event) != _selected.end())
					_selected.pop_back();
				else _selected.push_back(event);
				return(event);
			}
			else ePtr = ePtr->previous();
		}
	}
	return(0);
}

void SelectionState::revertExperiment()
{
	_experiment->revert();
}

void SelectionState::saveExperiment(const std::string &filename)
{
	if(_experiment) {
		if(filename.empty())
			_experiment->write(_filename + ".EDD");
		else {
			std::string::size_type pos = filename.find(".EDD");

			_filename = filename;
			if(pos == std::string::npos)
				pos = filename.find(".edd");
			if(pos != std::string::npos)
				_filename.erase(pos);
			_experiment->write(filename);
		}
	}
}

const SelectionList &SelectionState::selected() const
{
	return(_selected);
}

void SelectionState::selectFixation(const CPoint &clickPt, bool multiSelect)
{
	if(_experiment && _flags.test(EVENT_FIXATION)) {
		TrialData *trial = _experiment->trial(_trialPtr);
		FixationEvent *ePtr = trial->selectFixation(clickPt, _flags.test(EYE_LEFT), _flags.test(EYE_RIGHT));

		if(!multiSelect)                         //if user did not press CTRL key
			_selected.clear();                   //clear the previous selections

		if(ePtr) {                               //if a selection was made
			bool found = false;

			for(SelectionIter i = _selected.begin(); i != _selected.end(); ++i) {
				if(ePtr == *i) {                 //if selection was already selected, deselect it
					_selected.erase(i);
					found = true;
					break;
				}
			}
			if(!found)                           //otherwise add to selection list
				_selected.push_back(ePtr);
		}
	}
}

void SelectionState::initSelectFirst(){
	if(_experiment && _flags.test(EVENT_FIXATION) && ((_experiment->trial(_trialPtr))->hasLeftData() 
		|| (_experiment->trial(_trialPtr))->hasRightData() )) {
		TrialData *trial = _experiment->trial(_trialPtr);
		FixationEvent *ePtr = trial->selectFirstFixation(_flags.test(EYE_LEFT), _flags.test(EYE_RIGHT));
		_selected.push_back(ePtr);
	}
	return;
}

void SelectionState::setFlag(int flag, bool state)
{
	_flags.set(flag, state);

	if((flag == EVENT_FIXATION || flag == EYE_LEFT || flag == EYE_RIGHT) && state == false)
		_selected.clear();
}

void SelectionState::setLabelFontHt(int ht)
{
	_labelFontHt = ht;
}

void SelectionState::setJHookThreshold(int jh)
{
	_jhookThreshold = jh;
}

void SelectionState::setMultiChange(bool mc)
{
	 _useMultiChange = mc;
}

TrialData *SelectionState::trial()
{
	if(_experiment)
		return(_experiment->trial(_trialPtr));
	return(0);
}
