/*  Eyetracking software creates and executes experiments using 
    Eyelink II hardware and libraries by SR Research

    Copyright (C) 2004  David J. Stracuzzi, University of Massachusetts

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "StdAfx.h"
#include <cctype>
#include "Tokenizer.h"

#define USE_DISP_CHANGE_INFO 1

Tokenizer::Tokenizer()
  : _lineno(0), _fileLen(0), checkPosition(false)
{}

Tokenizer::~Tokenizer()
{
	_ifstrm.close();
}

void Tokenizer::clearFlags()
{
	_ifstrm.clear();
}

void Tokenizer::close()
{
	_ifstrm.close();
}

int Tokenizer::fileLength() const
{
	return(_fileLen);
}

int Tokenizer::filePosition()
{
	return(static_cast<int>(_ifstrm.tellg()));
}

bool Tokenizer::isOkay() const
{
	return(_ifstrm.is_open() && !_ifstrm.fail());
}

unsigned int Tokenizer::lineNumber() const
{
	return(_lineno);
}

void Tokenizer::open(const std::string &fileName)
{
	int start, end;

	_ifstrm.open(fileName.c_str(), std::ios::in);

	start = static_cast<int>(_ifstrm.tellg());
	_ifstrm.seekg(0, std::ios::end);
	end = static_cast<int>(_ifstrm.tellg());
	_ifstrm.seekg(0, std::ios::beg);

	_lineno = 0;
	_fileLen = end - start;
}

char Tokenizer::peekChar(bool ignoreWhitespace)
{
	if(ignoreWhitespace)
		readWhitespace();
	return(_ifstrm.peek());
}

char Tokenizer::readChar(bool ignoreWhitespace)
{
	char temp;

	if(ignoreWhitespace)
		readWhitespace();
	_ifstrm.get(temp);
	return(temp);
}

Tokenizer::EventType Tokenizer::readEventType()
{
	Tokenizer::EventType result = UNKNOWN_EVENT;

	while(!_ifstrm.fail() && result == UNKNOWN_EVENT) {
		std::string str;

		readWhitespace();                        //read any initial whitespace
		_ifstrm>>str;                            //read the event type

		if(str.compare("EBLINK") == 0)       result = BLINK_END;
		else if(str.compare("SBLINK") == 0)  result = BLINK_START;
		else if(str.compare("EFIX") == 0)    result = FIXATION;
		else if(str.compare("MSG") == 0)     result = MESSAGE;
		else if(str.compare("ESACC") == 0)   result = SACCADE;
		else if(str.compare("END") == 0)     result = TRIAL_END;
		else if(str.compare("SFIX") == 0)	{
			getline(_ifstrm, str, '\n');
			++_lineno;
			result = UNKNOWN_EVENT;
		}
		else if(str.compare("SSACC") == 0)	{
			getline(_ifstrm, str, '\n');
			++_lineno;
			result = UNKNOWN_EVENT;
		}
		else {					//didn't find one of these events		

			if(checkPosition && USE_DISP_CHANGE_INFO == 1){
				_lastX = readInteger();
				checkPosition = false;
				gotChangeCoords = true;
				
			}
			getline(_ifstrm, str, '\n');         //  read to EOL and try again
			++_lineno;	
		}
	}
	return(result);                              //file stream error or EOF
}

float Tokenizer::readFloat()
{
	float temp;

	readWhitespace();
	_ifstrm>>temp;
	return(temp);
}

int Tokenizer::readInteger()
{
	int temp;

	readWhitespace();
	_ifstrm>>temp;
	return(temp);
}

std::string Tokenizer::readLine()
{
	std::string str;

	getline(_ifstrm, str);
	++_lineno;
	return(str);
}

Tokenizer::MessageType Tokenizer::readMessageType()
{
	std::string str;

	readWhitespace();
	_ifstrm>>str;
	if(str.compare("DISPLAY") == 0) {            //DISPLAY COORDS message
		_ifstrm>>str;
		if(str.compare("COORDS") == 0)			 return(DISPLAY_SIZE);

		else if(str.compare("CHANGE") == 0){
			_ifstrm>>str;
			if (str.compare("STARTED") == 0)	return(DISP_CHANGE_START);
			else if(str.compare("COMPLETED") == 0)	return(DISP_CHANGE_END);
		}
	}
	else if(str.compare("ENDBUTTON") == 0)       return(END_BUTTON);
	else if(str.compare("GAZE") == 0) {          //GAZE TARGET ON message
		_ifstrm>>str;
		if(str.compare("TARGET") == 0) {
			_ifstrm>>str;
			if(str.compare("ON") == 0)           return(GAZE_TARGET_ON);
		}
	}
	else if (str.compare("LOADED") ==0){
		_ifstrm>>str;
		if(str.compare("IMAGE") == 0)
			return(IMAGE);
	}
	else if(str.compare("QUESTION_ANSWER") == 0) return(QSTN_ANSWER);
	else if(str.compare("REGION") == 0)          return(REGION);
	else if(str.compare("SYNCTIME") == 0)        return(SYNC_TIME);
	else if(str.compare("TRIAL") == 0)           return(TRIAL_STATUS);
	else if(str.compare("TRIALID") == 0)         return(TRIAL_ID);
	else if(str.compare("TRIAL_RESULT") == 0)    return(TRIAL_RESULT);
	else if(str.compare("TRIGGERED") == 0) {     //TRIGGERED FIXATION REGION message
		_ifstrm>>str;
		if(str.compare("FIXATION") == 0) {
			_ifstrm>>str;
			if(str.compare("REGION") == 0)		 return(TRIGGERED); // jdk change != -> ==
		}

	}
	else if(str.compare("EXPERIMENT") == 0)		return(EXPERIMENT_STATUS);
	else if(str.compare("DRIFTCORRECT") == 0)	return(DRIFTCORRECT);
	return(UNKNOWN_MESSAGE);
}

short Tokenizer::readShortInt()
{
	short temp;

	readWhitespace();
	_ifstrm>>temp;
	return(temp);
}

std::string Tokenizer::readString(int nChars)
{
	std::string str;

	if(nChars > 0) 
		for(int i=0; i < nChars; ++i)
			str.push_back(_ifstrm.get());
	else {
		readWhitespace();
		if(_ifstrm.peek() == '\'')
			getline(_ifstrm, str, '\'');
		else _ifstrm>>str;
	}
	return(str);
}

unsigned int Tokenizer::readUnsignedInt()
{
	unsigned int temp;

	readWhitespace();
	_ifstrm>>temp;
	return(temp);
}

void Tokenizer::readWhitespace()
{
	const int newline = '\n';
	int ch = _ifstrm.peek();

	while(_ifstrm.fail() == false && std::isspace(ch)) {
		if(ch == newline) ++_lineno;
		_ifstrm.ignore();
		ch = _ifstrm.peek();
	}
}

int Tokenizer::getLastX()
{
	return _lastX;
}

int Tokenizer::getLastY()
{
	return _lastY;
}

