/*  Eyetracking software creates and executes experiments using 
    Eyelink II hardware and libraries by SR Research

    Copyright (C) 2004  David J. Stracuzzi, University of Massachusetts

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "StdAfx.h"
#include "EyeDoctor.h"
#include "ExperimentData.h"
#include "SelectionState.h"
#include "TrackerView.h"

IMPLEMENT_DYNCREATE(CTrackerView, CScrollView)

CTrackerView::CTrackerView()
  : _bitmap(0), _state(0), _ctrlDown(false), _shiftDown(false)
{}

CTrackerView::~CTrackerView()
{
	if(_bitmap) delete _bitmap;
}

BEGIN_MESSAGE_MAP(CTrackerView, CScrollView)
	ON_WM_KEYDOWN()
	ON_WM_KEYUP()
	ON_WM_LBUTTONUP()
	ON_WM_RBUTTONUP()
END_MESSAGE_MAP()

#ifdef _DEBUG
void CTrackerView::AssertValid() const
{
	CScrollView::AssertValid();
}

void CTrackerView::Dump(CDumpContext& dc) const
{
	CScrollView::Dump(dc);
}
#endif //_DEBUG

void CTrackerView::ClearBitmap(COLORREF color)
{
	CBitmap *pOldBmp;
	CBrush *pOldBrush, brush(color);
	CSize size = GetTotalSize();
	CDC dc, *wndDC = GetDC();

	dc.CreateCompatibleDC(wndDC);                //create DC for updating bitmap
	ReleaseDC(wndDC);

	pOldBmp = dc.SelectObject(_bitmap);          //select the bitmap into context
	pOldBrush = dc.SelectObject(&brush);         //select brush

	dc.PatBlt(0, 0, size.cx, size.cy, PATCOPY);  //draw the background
	dc.SelectObject(pOldBmp);                    //deselect bitmap and brush
	dc.SelectObject(pOldBrush);
}

void CTrackerView::DrawBitmap()
{
	const SelectionList &selected = _state->selected();
	TrialData *trial = _state->trial();
	CBitmap *pBmp;
	CBrush *pBrush, bgBrush(_state->colors().background);
	CDC dc, *wndDC = GetDC();
	CFont labelFont, stimulusFont, *pFont;
	CSize size = GetTotalSize();

	dc.CreateCompatibleDC(wndDC);                //create context for bitmap
	dc.SetMapMode(MM_TEXT);                      //mapping mode most useful for text
	ReleaseDC(wndDC);                            //have bitmap context, release screen context

	labelFont.CreatePointFont(_state->labelFontHt()*10, "Courier New", &dc);
	stimulusFont.CreatePointFont(10*trial->fontHeight(), trial->fontName().c_str(), &dc);

	pBmp = dc.SelectObject(_bitmap);             //select the bitmap
	pBrush = dc.SelectObject(&bgBrush);          //select background brush
	pFont = dc.SelectObject(&stimulusFont);      //select the font

	dc.PatBlt(0, 0, size.cx, size.cy, PATCOPY);  //init the background
                                                 //draw the character regions
	trial->drawRegions(&dc, _state->colors(), _state->flags());

	trial->drawImages(&dc, _state->colors(), _state->flags());

	dc.SelectObject(&labelFont);                 //switch fonts
	                                             //draw events
	trial->drawEvents(&dc, _state->colors(), _state->flags());

	_state->setFlag(SELECTED, true);             //highlight selected fixation(s)
	for(constSelectionIter i = selected.begin(); i != selected.end(); ++i) {
		EyeEvent *event = *i;
		event->draw(&dc, _state->colors(), _state->flags());
	}
	_state->setFlag(SELECTED, false);

	dc.SelectObject(pBmp);                       //deselect local resources
	dc.SelectObject(pBrush);
	dc.SelectObject(pFont);
}

void CTrackerView::GetDPIFontMetrics(int &width, int &height)
{
	CDC dc, *wndDC = GetDC();
	CFont font;
	LOGFONT fontInfo;

	dc.CreateCompatibleDC(wndDC);
	dc.SetMapMode(MM_TEXT);
	ReleaseDC(wndDC);

	font.CreateFont(20, 0, 0, 0, FW_NORMAL, FALSE, FALSE, 0, ANSI_CHARSET,
					OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY,
					DEFAULT_PITCH | FF_DONTCARE, "Courier New");
	
	//font.CreatePointFont(200, "Courier New", &dc);
	font.GetLogFont(&fontInfo);
	width = fontInfo.lfWidth;
	height = fontInfo.lfHeight;
}

void CTrackerView::GenerateMenu(CMenu &menu, bool hasSelection, bool trialDeleted)
{
	std::string str = (trialDeleted ? "Undelete Trial" : "Delete Trial");
	int state = MF_STRING;

	menu.CreatePopupMenu();

	if(!hasSelection) state |= MF_GRAYED;

	menu.AppendMenu(state, ID_MERGE_FIXATION_PREV, "Merge Fixations (Prev)");
	menu.AppendMenu(state, ID_MERGE_FIXATION_NEXT, "Merge Fixations (Next)");
	menu.AppendMenu(state, ID_DELETE_FIXATION, "Delete Fixation");
	menu.AppendMenu(state, ID_IGNORE_FIXATION, "Ignore Fixation");
	menu.AppendMenu(MF_STRING, ID_DELETE_TRIAL, str.c_str());
	menu.AppendMenu(state, ID_TRUNCATE_TRIAL, "Truncate Trial");
}

void CTrackerView::OnDraw(CDC* pDC)
{
	CSize size = GetTotalSize();
	CBitmap *pBmp;
	CDC dc;

	if(!_bitmap) return;

	dc.CreateCompatibleDC(pDC);
	pBmp = dc.SelectObject(_bitmap);

	pDC->BitBlt(0, 0, size.cx, size.cy, &dc, 0, 0, SRCCOPY);

	dc.SelectObject(pBmp);
}

void CTrackerView::OnInitialUpdate()
{
	CSize sizeTotal;

	CScrollView::OnInitialUpdate();

	sizeTotal.cx = sizeTotal.cy = 100;
	SetScrollSizes(MM_TEXT, sizeTotal);

	
}

void CTrackerView::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags)
{
	if(_state->hasExperiment()) {
		const SelectionList &selected = _state->selected();
		TrialData *trial = _state->trial();
		bool update = true;

		switch(nChar) {
			case VK_CONTROL:
				_ctrlDown = true;
				update = false;
				break;
			case VK_SHIFT:
				_shiftDown = true;
				update = false;
				break;
			case VK_DOWN:
				for(constSelectionIter i = selected.begin(); i != selected.end(); ++i) {
					FixationEvent *event = *i;
					event->update(0, nRepCnt);
					event->update(trial->getRegionLocation(event->pixelLocation()));
				}
				break;
			case VK_LEFT:
				if(_shiftDown) {
					for(constSelectionIter i = selected.begin(); i != selected.end(); ++i) {
						FixationEvent *event = *i;
						event->update(-static_cast<int>(nRepCnt), 0);
						event->update(trial->getRegionLocation(event->pixelLocation()));
					}
				}
				else {
					for(UINT i = 0; i < nRepCnt; ++i)
						_state->prevFixation(_ctrlDown);
				}
				break;
			case VK_RIGHT:
				if(_shiftDown) {
					for(constSelectionIter i = selected.begin(); i != selected.end(); ++i) {
						FixationEvent *event = *i;
						event->update(nRepCnt, 0);
						event->update(trial->getRegionLocation(event->pixelLocation()));
					}
				}
				else {
					for(UINT i = 0; i < nRepCnt; ++i)
						_state->nextFixation(_ctrlDown);
				}
				break;
			case VK_UP:
				for(constSelectionIter i = selected.begin(); i != selected.end(); ++i) {
					FixationEvent *event = *i;
					event->update(0, -static_cast<int>(nRepCnt));
					event->update(trial->getRegionLocation(event->pixelLocation()));
				}
				break;
			default:
				UINT key = nChar;
				update = false;
		}

		if(update) theApp.UpdateAllViews();
	}

	CScrollView::OnKeyDown(nChar, nRepCnt, nFlags);
}

void CTrackerView::OnKeyUp(UINT nChar, UINT nRepCnt, UINT nFlags)
{
	if(_state->hasExperiment()) {
		if(nChar == VK_CONTROL)
			_ctrlDown = false;
		else if(nChar == VK_SHIFT)
			_shiftDown = false;
	}


	CScrollView::OnKeyUp(nChar, nRepCnt, nFlags);
}

void CTrackerView::OnLButtonUp(UINT nFlags, CPoint point)
{
	if(_state->hasExperiment()) {
		CPoint clickPt = GetScrollPosition();

		clickPt += point;
		_state->selectFixation(clickPt, _ctrlDown);
		_state->initSelectFirst();
		theApp.UpdateAllViews();

	}

	CScrollView::OnLButtonUp(nFlags, point);
}

void CTrackerView::OnRButtonUp(UINT nFlags, CPoint point)
{
	if(_state->hasExperiment()) {
		CMenu menu;
		CPoint scrnPt(point);
		CPoint clickPt = GetScrollPosition();
		int result;
		bool haveSelection = false;
		bool trialDeleted = _state->trial()->testFlag(TRIAL_DELETED);

		clickPt += point;
		_state->selectFixation(clickPt, false);
		theApp.UpdateAllViews();

		if(_state->currFixation()) 
			haveSelection = true;

		GenerateMenu(menu, haveSelection, trialDeleted);
		ClientToScreen(&scrnPt);
		result = menu.TrackPopupMenuEx(TPM_LEFTALIGN | TPM_TOPALIGN | TPM_RETURNCMD | TPM_RIGHTBUTTON, 
									   scrnPt.x, scrnPt.y, this, NULL);
		switch(result) {
			case ID_MERGE_FIXATION_PREV:
				_state->currFixation()->mergeWithFixation(true);
				break;
			case ID_MERGE_FIXATION_NEXT:
				_state->currFixation()->mergeWithFixation(false);
				break;
			case ID_DELETE_FIXATION:
				_state->currFixation()->mergeWithSaccade();
				break;
			case ID_IGNORE_FIXATION:
				_state->currFixation()->deleteEvent();
				_state->currFixation()->update(CPoint(_state->currFixation()->charLocation().x, 99));
				break;
			case ID_DELETE_TRIAL:
				_state->trial()->setFlag(TRIAL_DELETED, !trialDeleted);
				break;
			case ID_TRUNCATE_TRIAL:
				_state->currFixation()->truncateFutureEvents();
				break;
		}
		theApp.UpdateAllViews();
	}

	CScrollView::OnRButtonUp(nFlags, point);
}

void CTrackerView::SetDisplaySize(int width, int height)
{
	CSize size(width, height);
	CDC *pDC = GetDC();

	SetScrollSizes(MM_TEXT, size);

	if(_bitmap) delete _bitmap;
	_bitmap = new CBitmap();
	_bitmap->CreateDiscardableBitmap(pDC, width, height);
	ReleaseDC(pDC);

	ClearBitmap(0xFFFFFF);

	Invalidate();
	UpdateWindow();
}

void CTrackerView::SetSelectionState(SelectionState *state)
{
	_state = state;
}

void CTrackerView::UpdateView()
{
	if(_state->hasExperiment())
		DrawBitmap();
	else ClearBitmap(0xFFFFFF);
	Invalidate();
	UpdateWindow();
}
