/*  Eyetracking software creates and executes experiments using 
    Eyelink II hardware and libraries by SR Research

    Copyright (C) 2004  David J. Stracuzzi, University of Massachusetts

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "StdAfx.h"
#include "TrialData.h"
#include "BitmapFile.h"

#include <cmath>
#include <sstream>
#include <utility>
#define PROPER_MERGE
#define ROUND(x) ((short(x) - short((x)-0.5) == 0 ) ? short((x)+1) : short(x))

TrialData::TrialData()
  : _leftEvents(0), _rightEvents(0), _trialId(), _fontName("Courier New"), 
    _endTime(-1), _disp2Time(0), _disp3Time(0), _syncTime(-1), _fontHt(10), 
	_endButton(-1), _index(-1), _previewBdry(0), _previewResult(0), _qstnAnswer(-1), _images(0), _dispChangeCoords(CPoint(0,0)), dcOK(true), dcWithin5(true), dcWithin10(true), hasLongSacc(false)
{}

TrialData::TrialData(const std::string &trialId)
  : _leftEvents(0), _rightEvents(0), _trialId(trialId), _fontName("Courier New"), 
    _endTime(-1), _disp2Time(0), _disp3Time(0), _syncTime(-1), _fontHt(10), 
	_endButton(-1), _index(-1), _previewBdry(0), _previewResult(0), _qstnAnswer(-1), _images(0), _dispChangeCoords(CPoint(0,0)), dcOK(true), dcWithin5(true), dcWithin10(true), hasLongSacc(false)
{}

TrialData::~TrialData()
{
	while(_leftEvents) {                         //delete left eye events
		EyeEvent *temp = _leftEvents;
		_leftEvents = _leftEvents->next(false);
		delete temp;
	}

	while(_rightEvents) {                        //delete right eye events
		EyeEvent *temp = _rightEvents;
		_rightEvents = _rightEvents->next(false);
		delete temp;
	}

 	while(_regions.empty() == false) {           //delete character regions
		TrackRegion *tr = _regions.front();
		_regions.pop_front();
		delete tr;
	}
}

void TrialData::drawEvents(CDC *pDC, const Colors &colors, const StateFlags &flags)
{
	EyeEvent *evtPtr = _leftEvents;

	if(flags.test(EYE_LEFT)) {
		while(evtPtr) {
			evtPtr->draw(pDC, colors, flags);
			evtPtr = evtPtr->next(false);
		}
	}

	if(flags.test(EYE_RIGHT)) {
		evtPtr = _rightEvents;
		while(evtPtr) {
			evtPtr->draw(pDC, colors, flags);
			evtPtr = evtPtr->next(false);
		}
	}
}

void TrialData::drawRegions(CDC *pDC, const Colors &colors, const StateFlags &flags)
{
	for(TRIter i = _regions.begin(); i != _regions.end(); ++i) {
		TrackRegion *tr = *i;
		tr->draw(pDC, colors, flags);
	}
}

void TrialData::drawImages(CDC *pDC, const Colors &colors, const StateFlags &flags)
{
		pDC->SetBkMode(TRANSPARENT);
		CSize zero(0,0);
		CSize size(_displayCoords.Width(), _displayCoords.Height());
		pDC->SetTextColor(colors.stimulus);

		std::vector<IMAGEPOS>::iterator image = _images.begin();
		while(image != _images.end()){
			CBitmapFile smallBmp;
			CDC small_memdc;

			smallBmp.CreateCompatibleBitmap(pDC, image->upperLeft.x, image->upperLeft.y);
			small_memdc.CreateCompatibleDC(pDC);


			if(image->upperLeft.x == 0 && image->upperLeft.y == 0){
				smallBmp.LoadBitmap(image->fileName.c_str(), size);
			}
			else
				smallBmp.LoadBitmap(image->fileName.c_str(), zero);

			small_memdc.SelectObject(smallBmp);
			pDC->BitBlt(image->upperLeft.x, image->upperLeft.y, _displayCoords.Width(), _displayCoords.Height(),&small_memdc, 0,0,SRCCOPY);
					
			smallBmp.DeleteObject();
			small_memdc.DeleteDC();
			image++;
		}
}

unsigned int TrialData::duration() const
{
	return(_endTime);
}

short TrialData::endButton() const
{
	return(_endButton);
}

void TrialData::export(std::ostream &exportStrm, std::ostream &changeStrm, const ExportParams &param)
{
	std::istringstream is(_trialId.idString());
	std::list<FixationEvent> exportEvents;       //list of *copied* events
	std::list<FixationEvent>::iterator i1, i2;   //pointers into the list
	EyeEvent *event = param.useRight ? _rightEvents : _leftEvents;
	int cond, item;
	char c;

	if(event == 0)                               //trial has no writable events, skip it
		return;

	is>>c>>cond;                                 //get trial condition and item
	is>>c>>item;

	exportStrm<<_index<<' '<<cond<<' '<<item<<' ';
	exportStrm<<duration()<<' '<<_endButton<<' ';
	exportStrm<<_previewBdry<<' '<<_previewResult<<' ';

	while(event) {                               //make a list of all fixations (copied)
		FixationEvent *fEvt = dynamic_cast<FixationEvent*>(event);
		if(fEvt && !fEvt->isDeleted())
			exportEvents.push_back(*fEvt);
		event = event->next(false);
	}

	i1 = i2 = exportEvents.begin();              //perform final fix procedure
	++i2;
	while(i1 != exportEvents.end() && i2 != exportEvents.end()) {
		int dist = i1->charLocation().x - i2->charLocation().x;

		if(dist < 0) dist = -dist;

		if(dist <= param.mergeDist) {            //try to merge short, nearby events
			if(i1->duration() < param.mergeLen) {
				EyeEvent *tmp = &(*i2);
				int timeDiff = static_cast<int>(i2->startTime() - i1->startTime());

				changeStrm<<"Merged "<<(i1->isRightEye() ? "right" : "left");
				changeStrm<<" eye fixation "<<i1->index();
				changeStrm<<" into fixation "<<i2->index()<<"\r\n";
				tmp->update(-timeDiff, 0);
				exportEvents.erase(i1);
				i1 = i2;
			}
			else if(i2->duration() < param.mergeLen) {
				EyeEvent *tmp = &(*i1);

				changeStrm<<"Merged "<<(i2->isRightEye() ? "right" : "left");
				changeStrm<<" eye fixation "<<i2->index();
				changeStrm<<" into fixation "<<i1->index()<<"\r\n";
				tmp->update(0, i2->endTime() - i1->endTime());
				exportEvents.erase(i2);
				i2 = i1;
			}
#ifdef PROPER_MERGE
			else i1=i2;                // we met the distance criterion, but not ^M
                                                // the duration criterion, so increment iterator^M
                                                // but don't perform a merge^M
#endif
		}
		else if(dist <= param.deleteDist) {      //delete very short, somewhat nearby events
			if(i1->duration() < param.deleteLen) {
				changeStrm<<"Deleted "<<(i1->isRightEye() ? "right" : "left");
				changeStrm<<" eye fixation "<<i1->index()<<"\r\n";
				exportEvents.erase(i1);
				i1 = i2;
			}
			else if(i2->duration() < param.deleteLen) {
				changeStrm<<"Deleted "<<(i2->isRightEye() ? "right" : "left");
				changeStrm<<" eye fixation "<<i2->index()<<"\r\n";
				exportEvents.erase(i2);
				i2 = i1;
			}
#ifdef PROPER_MERGE
			else {
				i1 = i2;	// we met the distance criterion, but not the^M
                                                        // duration criterion, so don't perform any^M
			}                                        // deletion, but increment the iterator^M
#endif
		}
		else {
			i1 = i2;
		}
		++i2;
	}
                                                 //finally write number of events and events 
	exportStrm<<static_cast<int>(exportEvents.size());
	for(i1 = exportEvents.begin(); i1 != exportEvents.end(); ++i1)
		i1->export(exportStrm);
	exportStrm<<std::endl;
}

int TrialData::fontHeight() const
{
	return(_fontHt);
}

const std::string &TrialData::fontName() const
{
	return(_fontName);
}

CPoint TrialData::getRegionLocation(const CPoint &point) const
{
	for(constTRIter i = _regions.begin(); i != _regions.end(); ++i) {
		const CRect &rect = (*i)->boundary();
		if(rect.PtInRect(point)) return((*i)->location());
	}
	return(CPoint(-1, -1));
}

bool TrialData::hasLeftData() const
{
	return(_leftEvents != 0);
}

bool TrialData::hasRightData() const
{
	return(_rightEvents != 0);
}

bool TrialData::isDependent() const
{	
	return _trialId.isDependent();
	/*
	std::istringstream os(_trialId.idString());
	int cond;

	os.ignore();
	os>>cond;
	return(cond > 99); */
}

void TrialData::import(Tokenizer &tok, const int j_hook_size) throw(std::exception)
{
	unsigned int _disp2TimeStart = 0;
	EyeEvent *blink = 0, *blinkEnd = 0;
	EyeEvent *leftEnd = 0, *rightEnd = 0;
	std::string buf, buf2, status;
	float x1, y1, x2, y2, dud;
	unsigned int end, start, startTime = 0;
	short index, len;
	bool isBlink = false, isRight, done = false, gotELend = false, gotETend = false;
	bool endBlink = false;
	IMAGEPOS ip;
	CPoint pt;
	bool checkedDispChange = false;

	tok.checkPosition = false;
	tok.gotChangeCoords = false;

	_flags.set(BLINK_CHANGE_ERROR, false);
	_flags.set(J_HOOK_DISP_CHANGE, false);
	FixationEvent *lastEvent = 0;

	do {
		//if (tok.gotChangeCoords && !tok.checkPosition && checkedDispChange){
		//	this->setFlag(ERROR, true);
		//}



		EyeEvent *ePtr = 0;

		if (/*_dispChangeCoords.x == 0 && */ _disp2TimeStart >0 && tok.gotChangeCoords){
			_dispChangeCoords.x = tok.getLastX();
			_dispChangeCoords.y = 0;

		/*	if(lastEvent == 0)
				_dcVector.push_back(dcData(_dispChangeCoords, _disp2TimeStart,0, 0, 0));
			else{
			  _dcVector.push_back(dcData(_dispChangeCoords, _disp2TimeStart, lastEvent->startTime(), lastEvent->startTime(), lastEvent->endTime()));
			}//_dcVector.push_back(dcData(_dispChangeCoords, _disp2TimeStart));
*/
			tok.gotChangeCoords = false;
			tok.checkPosition = false;
		}
		switch(tok.readEventType()) {
		case Tokenizer::BLINK_START:            //starts a blink (ends with saccade)
			isBlink = true;
			endBlink = false;
			//tok.readChar(); // eat left/right
			//tok.readInteger();// eat timestamp
			break;
		case Tokenizer::BLINK_END:
			//tok.readChar();//eat left/right
			//tok.readInteger(); //eat start timestamp
			//tok.readInteger();// eat end timestamp
			endBlink = true;
			tok.checkPosition = true;
			break;
		case Tokenizer::FIXATION:                //fixation event
			if(tok.readChar() == 'L') isRight = false;
			else isRight = true;

			start = tok.readUnsignedInt();
			end = tok.readUnsignedInt();
			tok.readUnsignedInt();
			x1 = tok.readFloat();
			y1 = tok.readFloat();
			dud = tok.readInteger();

			if(tok.isOkay() && start > _syncTime && end < _endTime) {
				// start change:jdk
				FixationEvent *f;	
				if(0 < _disp2Time && _disp2Time < (start - startTime ))
					f = new FixationEvent(isRight, false, true, start, end, 
													ROUND(x1), ROUND(y1));
				else 
					f = new FixationEvent(isRight, false, false, start, end, 
													 ROUND(x1), ROUND(y1));
				// end 

				// FixationEvent *f = new FixationEvent(isRight, false, false, start, end, 
				//									 ROUND(x1), ROUND(y1));

				CPoint loc = getRegionLocation(f->pixelLocation());
				f->updateInitial(loc);                  //translate pixels to chars
				if(_disp2Time > 0 && !checkedDispChange){
					checkedDispChange = true;
					if(f->pixelLocation().x < (_dispChangeCoords.x-j_hook_size))
						this->setFlag(J_HOOK_DISP_CHANGE,true);
				}
				ePtr = f;
				lastEvent = f;
				if ((start - startTime) < _disp2Time && _disp2Time <= (end - startTime)){
					//_dcVector.pop_back();
					if(lastEvent){
						_dcVector.push_back(dcData(_dispChangeCoords, _disp2TimeStart, lastEvent->startTime(), start-startTime/*lastEvent->leadSaccade()->startTime()*/, end-startTime/*lastEvent->leadSaccade()->endTime()*/));
				
					}
					else {  // otherwise, trial recording began during this saccade, and change triggered during first saccade 
							// so load default data for last fix:
						_dcVector.push_back(dcData(_dispChangeCoords, _disp2TimeStart, start-startTime, start-startTime/*lastEvent->leadSaccade()->startTime()*/, end-startTime/*lastEvent->leadSaccade()->endTime()*/));
					}

				}

			}
			else                                 //set index counters to zero
				FixationEvent::initialize(isRight);
			break;
		case Tokenizer::MESSAGE:                 //message event
			start = tok.readUnsignedInt();

			switch(tok.readMessageType()) {      //process message
			case Tokenizer::FONT:
				_fontName = tok.readString();
				_fontHt = tok.readShortInt();
				break;
			case Tokenizer::GAZE_TARGET_ON:
				startTime = start;

				break;
			case Tokenizer::QSTN_ANSWER:
				_qstnAnswer = tok.readInteger();
				break;
			case Tokenizer::IMAGE:
				tok.readInteger();		// read display 1 or display 2
				ip.fileName = tok.readString();
				if (!strstr(tok.readString().c_str(), "AT")){
					ip.upperLeft = CPoint(0,0);
					_images.push_back(ip);
					break;
				}// read "AT"
				tok.readChar();			// read "(";			
				pt.x = tok.readInteger();
				tok.readChar();			// read ","
				pt.y = tok.readInteger();
				tok.readChar();			// read ")"
				ip.upperLeft = pt;
				_images.push_back(ip);
				break;

			case Tokenizer::REGION:
				tok.readString();
				index = tok.readShortInt();
				len = tok.readShortInt();
				tok.readChar(false);             //read one leading white space
				//buf = tok.readString(len);       //read region text


				// kludge to fix misrecorded length data for 
				// multi-byte characters
				if(len==1){
					buf = tok.readString(1);
					if(_ismbblead(buf.at(0))){
						buf2 = tok.readString(1);
						buf.append(buf2);
					}
				}

				else {
					buf = tok.readString(len);
				}

				tok.readChar(false);             //read trailing white space
				x1 = tok.readShortInt();
				y1 = tok.readShortInt();
				x2 = tok.readShortInt();
				y2 = tok.readShortInt();

				if(tok.isOkay()) {               //read okay, set the character indicies
					TrackRegion *region = new TrackRegion(ROUND(x1), ROUND(y1),
														  ROUND(x2), ROUND(y2), 
														  index, buf.c_str());
					CRect bdry = region->boundary();
					CPoint charLoc;

					if(_regions.empty())         //first region
						charLoc.SetPoint(0, 0);
					else {                       //subsequent regions
						int oldTop = _regions.back()->boundary().top;

						charLoc = _regions.back()->location();
						++charLoc.x;
						while(oldTop < bdry.top) {
							//oldTop += bdry.Height();
							oldTop = bdry.top;
							charLoc.x = 0;
							++charLoc.y;
						}
					}
					region->setLocation(charLoc);
					_regions.push_back(region);
				}
				break;
			case Tokenizer::SYNC_TIME:           //set SYNCTIME
				if(startTime == 0) startTime = start;
				_syncTime = start;
				break;
			case Tokenizer::TRIAL_RESULT:        //set endButton and endTime
				_endTime = start;
				_endButton = tok.readShortInt();
				if(_endButton == 0) 
					_flags.set(TRIAL_TIMEOUT, true);
				gotETend = true;
				break;
			case Tokenizer::DISP_CHANGE_START:
				//_disp2Time = start;
				tok.checkPosition = true;
				_disp2TimeStart = start - startTime + DISPLAY_REFRESH_LAG;
				if(isBlink && !endBlink){
					_flags.set(BLINK_CHANGE_ERROR, true);
					tok.checkPosition = false; // wait until end blink to checkPosition
												// if we get the change in a blink state
				}
				//_dispChangeCoords = CPoint(tok.getLastX(), tok.getLastY());
				break;
			case Tokenizer::DISP_CHANGE_END:
				//tok.checkPosition = true;
				//_dispChangeCoords.x = tok.getLastX();
				_disp2Time = start - startTime + DISPLAY_REFRESH_LAG; // jdk: add 5 ms to approximate screen refresh
				break;
			case Tokenizer::EXPERIMENT_STATUS:
				status = tok.readString();
				if(status.compare("ABORTED") == 0){
					gotETend = true;
					gotELend = true;
					_flags.set(TRIAL_ABORTED, true);
				}
				break;
			case Tokenizer::TRIAL_STATUS:
				 status = tok.readString();
				 if(status.compare("REPEATED") == 0){
					 gotETend = true;
					 gotELend = true;
					 done = true;
					 _flags.set(TRIAL_REPEATED, true);
					 break;
				 }
				 if(status.compare("ABORTED") == 0 || status.compare("ERROR") == 0 ){
                         gotETend = true;
						 if (status.compare("ERROR") == 0){
							 _flags.set(TRIAL_ERROR, true);
							 gotELend = true; // kludge to skip trial errors
						 }
						 else {
							 _flags.set(TRIAL_ABORTED, true);
							 gotELend = true; 
						 }
						 break;
				 }
				 if(status.compare("OK") == 0){
					 gotETend =true;
					 gotELend = true;
					 done = true;
					 _flags.set(TRIAL_NO_END_MESSAGE, true); 
					 _flags.set(TRIAL_ERROR, true);	// kludge to catch EyeTrack Error
													// if we get here, then there
													// was actually a trial ERROR and 
													// no data was recorded, so the "END"
													// and "TRIAL_RESULT" flags will not
													// be found, ie, next message will
													// be the start of the next TRIAL_ID
				 }

                break;
			case Tokenizer::TRIAL_ID:
				gotETend = true;
				gotELend = true;
				_flags.set(TRIAL_ERROR, true);
				break;
			case Tokenizer::DRIFTCORRECT:
				tok.readChar();  // EAT which eye
				status = tok.readString();
				if(status.compare("ABORTED") == 0){
					gotETend = true;
					gotELend = true;
					_flags.set(TRIAL_ABORTED, true);
				}
				break;
			}                                   //end MESSAGE processing section
			break;
		case Tokenizer::SACCADE:                 //Saccade event
			//tok.checkPosition = true;
			if(tok.readChar() == 'R') isRight = true;
			else isRight = false;
			start = tok.readUnsignedInt();
			end = tok.readUnsignedInt();
			tok.readUnsignedInt();

			if(tok.isOkay() == false) break;

			x1 = tok.readFloat();
			y1 = tok.readFloat();
			x2 = tok.readFloat();
			y2 = tok.readFloat();

			if(!tok.isOkay()) {
				x1 = x2 = y1 = y2 = -1;
				tok.clearFlags();
			}

			if(start > _syncTime && end < _endTime){
				if(0 < _disp2Time && _disp2Time < (start - startTime )){
					ePtr = new SaccadeEvent(isRight, isBlink, true, start, end, 
										ROUND(x1), ROUND(y1), ROUND(x2), ROUND(y2));  
				}
				else {
					ePtr = new SaccadeEvent(isRight, isBlink, false, start, end, 
										ROUND(x1), ROUND(y1), ROUND(x2), ROUND(y2));
				}
				if ((start - startTime) < _disp2Time && _disp2Time <= (end - startTime)){
					//_dcVector.pop_back();
					if(lastEvent){
						_dcVector.push_back(dcData(_dispChangeCoords, _disp2TimeStart, lastEvent->startTime(), start-startTime/*lastEvent->leadSaccade()->startTime()*/, end-startTime/*lastEvent->leadSaccade()->endTime()*/));
				
					}
					else {  // otherwise, trial recording began during this saccade, and change triggered during first saccade 
							// so load default data for last fix:
						_dcVector.push_back(dcData(_dispChangeCoords, _disp2TimeStart, start-startTime, start-startTime/*lastEvent->leadSaccade()->startTime()*/, end-startTime/*lastEvent->leadSaccade()->endTime()*/));
					}

				}
				else{

					//_dcVector.push_back(dcData(_dispChangeCoords, _disp2TimeStart, 0, start-startTime/*lastEvent->leadSaccade()->startTime()*/, end-startTime/*lastEvent->leadSaccade()->endTime()*/));
					//_dcVector.push_back(dcData(_dispChangeCoords, _disp2TimeStart, 0, 44, 88));
				}


			}

			isBlink = false;
			break;
		case Tokenizer::TRIAL_END:
			if(_endTime == static_cast<unsigned int>(-1)) {
				_endButton = 0;
				_endTime = tok.readUnsignedInt();
			}
			done = true;
			gotELend = true;
			break;

		default:
			break;
//			return;
		}

	

		if(ePtr) {                               //do some bookkeeping on events
			ePtr->updateInitial(-int(startTime), -int(startTime)); //scale event times
			if(ePtr->isRightEye()) {             //place event on proper list
				if(_rightEvents) rightEnd->setNext(ePtr);
				else _rightEvents = ePtr;
				rightEnd = ePtr;
			}
			else {
				if(_leftEvents) leftEnd->setNext(ePtr);
				else _leftEvents = ePtr;
				leftEnd = ePtr;
			}
		
		}
	} while(!(gotETend && gotELend));/*while(!done);*/

	for(EyeEvent *e = _leftEvents; e != 0; e = e->next(false)) {
		FixationEvent *fe = dynamic_cast<FixationEvent*>(e);
		if(fe) fe->testBlink();
	}

	for(EyeEvent *e = _rightEvents; e != 0; e = e->next(false)) {
		FixationEvent *fe = dynamic_cast<FixationEvent*>(e);
		if(fe) fe->testBlink();
	}

	_syncTime -= startTime;                      //adjust trial SYNC and END times
	_endTime -= startTime;
}

void TrialData::importData(Tokenizer &tok, int firstLine) throw(std::exception)
{
	const int height = (8*_fontHt)/7;
	const int width = (2*height)/3;
    const int x0 = static_cast<int>(1.5*static_cast<double>(width));
	const int y0 = static_cast<int>(2.5*static_cast<double>(height));
	constTRIter iter = _regions.begin();
	FixationEvent *lastEvent = 0;
	int nFixations;

	FixationEvent::initialize(true);             //init fixation counter
	_qstnAnswer = 1;                             //set default question answer

	_endTime = tok.readUnsignedInt();            //read the trial duration
	_endButton = tok.readShortInt();             //read the ending response 

	_previewBdry = tok.readShortInt();           //read rt2 (a-la Chuck)
	_previewResult = tok.readShortInt();         //read resp2  "    "
	_disp2Time = tok.readUnsignedInt();          //read extim1 "    "
	_disp3Time = tok.readUnsignedInt();          //read extim2 "    "
	nFixations = tok.readInteger();              //read number of fixations

	if(_previewBdry > 0) {
		for(int i=0; i < _previewBdry; ++i, ++iter);
		if(iter != _regions.end()) {
			TrackRegion *tr = *iter;
			tr->setAsBoundary();
		}
		else {
			std::ostringstream os;
			os<<"Error loading trial "<<_trialId.idString();
			os<<":\nBad preview region boundary location."<<std::endl;
			throw(std::exception(os.str().c_str()));
		}
	}

	for(int i=0; i < nFixations; ++i) {          //read each fixation, create matching saccades
		FixationEvent *f;
		CPoint charLoc, fixPt;
		unsigned int start, end;
		bool isBndry = false;

		charLoc.x = tok.readInteger();           //read character location and time
		charLoc.y = tok.readInteger() - firstLine;
		start = tok.readUnsignedInt();
		end = tok.readUnsignedInt();
                                                 //check if the event overlaps a display change
		if(_disp2Time > 0 && start <= _disp2Time && end >= _disp2Time)
			isBndry = true;
		if(_disp3Time > 0 && start <= _disp3Time && end >= _disp3Time)
			isBndry = true;

		fixPt.x = x0 + charLoc.x * width;
		fixPt.y = y0 + charLoc.y * height;

		f = new FixationEvent(true, false, isBndry, start, end, short(fixPt.x), short(fixPt.y));
		f->update(charLoc);

		isBndry = false;
		if(lastEvent) {                          //create saccade between prev and current fixation
			const CPoint &startPt = lastEvent->pixelLocation();
			SaccadeEvent *s; 
			
			if(_disp2Time > 0 && lastEvent->endTime() <= _disp2Time && start >= _disp2Time)
				isBndry = true;
			if(_disp2Time > 0 && lastEvent->endTime() <= _disp3Time && start >= _disp3Time)
				isBndry = true;

			s = new SaccadeEvent(true, false, isBndry, lastEvent->endTime(), start, 
			  				     short(startPt.x), short(startPt.y), 
								 short(fixPt.x), short(fixPt.y));
			s->setNext(f);
			lastEvent->setNext(s);
			lastEvent = f;
		}
		else 
			_rightEvents = lastEvent = f;
	}
}

void TrialData::importSen(Tokenizer &tok) throw(std::exception)
{
	const int height = 30;
	const int width = 20;
	TrackRegion *rgn;
	CPoint charLoc(-1, 0);
	int index = 0;
	short x = 0, y = 2*height;
	char buf[3];

	_fontHt = 7*height/8;
	buf[1] = '\0';

	while(tok.isOkay() && tok.peekChar(false) != '{') {
		buf[0] = tok.readChar(false);
//		buf[1] = tok.peekChar(false);
//		buf[2] = '\n';
//		if(_mbstrlen(buf) == 1){
//		}
//		else buf[1] = '\n';
		if(buf[0] == '\n') {
			charLoc.x = -1;
			++charLoc.y;
			x = 0;
			y += height;
		}
		else {
			++charLoc.x;
			x += width;
			rgn = new TrackRegion(x, y, x+width, y+height, index++, buf);
			rgn->setLocation(charLoc);
			_regions.push_back(rgn);
		}
	}
}

short TrialData::index() const
{
	return(_index);
}

short TrialData::questionStatus() const
{
	if(_qstnAnswer < 0) return(0);
	else return(_endButton == _qstnAnswer ? 1 : -1);
}

void TrialData::read(std::istream &is, int versionNum)
{
	std::string stmp;
	EyeEvent *last = 0;
	int nObjects = 0;

	is>>_flags;                                  //read trial flags
	getline(is, stmp, '\0');                     //read trial id
	_trialId = TrialId(stmp);

	if(versionNum == 2) {                        //V 0.2.0 only
		is.read(reinterpret_cast<char*>(&_endTime), sizeof(unsigned int));
		is.read(reinterpret_cast<char*>(&_syncTime), sizeof(unsigned int));
		is.read(reinterpret_cast<char*>(&_endButton), sizeof(short));
	}
	else {
		int rdSize = 4*sizeof(unsigned int) + sizeof(int) + 4*sizeof(short);

		getline(is, _fontName, '\0');            //read font name
		if(versionNum > 3)
			rdSize += sizeof(short);
		is.read(reinterpret_cast<char*>(&_endTime), rdSize);
	}
	                                             //All versions read events
	is.read(reinterpret_cast<char*>(&nObjects), sizeof(int));
	for(int i = 0; i < nObjects; ++i) {          //read left events
		EyeEvent *event = 0;
		char type;

		is.get(type);
		if(type == 'F') event = new FixationEvent();
		else if(type == 'S') event = new SaccadeEvent();
		else throw(std::exception("Error reading trial data: Bad left eye event\n"));
		event->read(is, versionNum);

		if(_leftEvents == 0) _leftEvents = event;
		else last->setNext(event);
		last = event;
	}

	last = 0;
	is.read(reinterpret_cast<char*>(&nObjects), sizeof(int));
	for(int i = 0; i < nObjects; ++i) {          //read right events
		EyeEvent *event = 0;
		char type;

		is.get(type);
		if(type == 'F') event = new FixationEvent();
		else if(type == 'S') event = new SaccadeEvent();
		else throw(std::exception("Error reading trial data: Bad right eye event\n"));
		event->read(is, versionNum);

		if(_rightEvents == 0) _rightEvents = event;
		else last->setNext(event);
		last = event;
	}

	is.read(reinterpret_cast<char*>(&nObjects), sizeof(int));
	for(int i = 0; i < nObjects; ++i) {
		TrackRegion *tr = new TrackRegion;
		tr->read(is, versionNum);
		_regions.push_back(tr);
	}

	is.read(reinterpret_cast<char*>(&nObjects), sizeof(int));
	for(int i = 0; i < nObjects; ++i) {
		dcData *dc = new dcData();
		dc->read(is, versionNum);
		_dcVector.push_back(dcData(dc->loc, dc->time, dc->fixTime, dc->leadStart, dc->leadEnd));
		delete(dc);
		//HERE
	}
}

/*dcData TrialData::readDcData(std::istream &is, int versionNum){
	int sz = sizeof(CRect) + sizeof(CPoint) + sizeof(short);
	std::ios::pos_type pos;
	short len = 1;

	pos = is.tellg();                            //mark current position in stream
	while(is.get() != '\0') ++len;               //find the end (length) of text string

	is.seekg(pos);                               //rewind to beginning of text
	_text = new char[len];
	is.read(_text, len);  
}*/

/*void TrialData::writeDcData(std::ostream &os) const{
	int sz = sizeof(CRect) + sizeof(CPoint) + sizeof(short) + sizeof(bool);
	short len = 0;

	while(_text[len] != '\0') ++len;             //compute length of string
                                                 //write region data
	os.write(reinterpret_cast<const char*>(&_boundary), sz);
	os.write(_text, len+1);
}*/

void TrialData::revert()
{
	EyeEvent *ePtr = _leftEvents;

	while(ePtr) {
		ePtr->revert();
		ePtr = ePtr->next(false);
	}

	ePtr = _rightEvents;
	while(ePtr) {
		ePtr->revert();
		ePtr = ePtr->next(false);
	}
	_flags.set(TRIAL_DELETED, false);
}

FixationEvent *TrialData::selectFixation(const CPoint &clickPt, bool leftEye, bool rightEye)
{
	EyeEvent *ePtr = _leftEvents;

	if(leftEye) {
		while(ePtr) {
			FixationEvent *fPtr = dynamic_cast<FixationEvent*>(ePtr);
			if(fPtr) {
				CRect loc = fPtr->rectLocation();
				if(loc.PtInRect(clickPt)) return(fPtr);
			}
			ePtr = ePtr->next();
		}
	}

	if(rightEye) {
		ePtr = _rightEvents;
		while(ePtr) {
			FixationEvent *fPtr = dynamic_cast<FixationEvent*>(ePtr);
			if(fPtr) {
				CRect loc = fPtr->rectLocation();
				if(loc.PtInRect(clickPt)) return(fPtr);
			}
			ePtr = ePtr->next();
		}
	}
	return(0);
}

FixationEvent *TrialData::selectFirstFixation(bool leftEye, bool rightEye)
{
	EyeEvent *ePtr = _leftEvents;

	if(leftEye) {
 		while(ePtr) {
			FixationEvent *fPtr = dynamic_cast<FixationEvent*>(ePtr);
			if(fPtr) {
				CRect loc = fPtr->rectLocation();
				return (fPtr);
				//if(loc.PtInRect(clickPt)) return(fPtr);
			}
			ePtr = ePtr->next();
		}
	}

	if(rightEye) {
		ePtr = _rightEvents;
		while(ePtr) {
			FixationEvent *fPtr = dynamic_cast<FixationEvent*>(ePtr);
			if(fPtr) {
				CRect loc = fPtr->rectLocation();
				//if(loc.PtInRect(clickPt)) return(fPtr);
				return(fPtr);
			}
			ePtr = ePtr->next();
		}
	}
	return(0);
}

void TrialData::setFlag(int flag_id, bool value)
{
	_flags.set(flag_id, value);
}

void TrialData::setIndex(short index)
{
	_index = index;
}

bool TrialData::testFlag(int flag_id) const
{
	return(_flags.test(flag_id));
}

const TrialId &TrialData::trialID() const
{
	return(_trialId);
}

void TrialData::write(std::ostream &os) const
{ 
	EyeEvent *ePtr = _leftEvents;
	std::string id = _trialId.idString();
	int writeSz = 4*sizeof(unsigned int) + sizeof(int) + 5*sizeof(short);
	int nObjs = 0;

	os<<_flags;                                  //write trial data
	os.write(id.c_str(), static_cast<int>(id.size()+1));
	os.write(_fontName.c_str(), static_cast<int>(_fontName.size()+1));
	os.write(reinterpret_cast<const char*>(&_endTime), writeSz);
 
	if(_leftEvents) nObjs = _leftEvents->length();
	else nObjs = 0;
	os.write(reinterpret_cast<char*>(&nObjs), sizeof(int));

	while(ePtr) {                                //write left eye events
		ePtr->write(os);
		ePtr = ePtr->next(false);
	}

	if(_rightEvents) nObjs = _rightEvents->length();
	else nObjs = 0;
	os.write(reinterpret_cast<char*>(&nObjs), sizeof(int));

	ePtr = _rightEvents;
	while(ePtr) {                                //write right eye events
		ePtr->write(os);
		ePtr = ePtr->next(false);
	}
                                                 //write track regions
	nObjs = static_cast<int>(_regions.size());
	os.write(reinterpret_cast<char*>(&nObjs), sizeof(int));
	for(constTRIter i = _regions.begin(); i != _regions.end(); ++i) {
		const TrackRegion *tr = *i;
		tr->write(os);
	}


	nObjs = static_cast<int>(_dcVector.size());
	os.write(reinterpret_cast<char*>(&nObjs), sizeof(int));
	typedef std::vector<DCDATA>::const_iterator dcIter;
	for(dcIter i = _dcVector.begin(); i != _dcVector.end(); ++i) {
		const dcData dc = *i;
		dc.write(os);
	}
}

void TrialData::writeSummary(std::ostream &os, bool tabbed) /*const*/
{
	EyeEvent *ePtr = _leftEvents;
	float meanFixDurL = 0.0, meanFixDurR = 0.0;
	float meanSaccDurL = 0.0, meanSaccDurR = 0.0;
	float varFixDurL = 0.0, varFixDurR = 0.0;
	float varSaccDurL = 0.0, varSaccDurR = 0.0;
	unsigned int saccEnd = 0, saccStart = 0;
	int fixIdx = -1;
	int nFixL = 0, nFixR = 0;
	int nSaccL = 0, nSaccR = 0;
	int nBlinkL = 0, nBlinkR = 0;
	/*bool dcOK = false;
	bool dcWithin5 = false;
	bool dcWithin10 = false; */
	unsigned int startFixPostDispChange = 0, endFixPreDispChange = 0, endFixPostDispChange = 0;
	dcLeadSaccStart = 0;
	dcLeadSaccEnd = 0;

	while(ePtr) {                                //compute left mean fixn & sacc durations
		const FixationEvent *fEvt = dynamic_cast<FixationEvent*>(ePtr);
		const SaccadeEvent *sEvt = dynamic_cast<SaccadeEvent*>(ePtr);

		if(fEvt) {
			meanFixDurL += fEvt->duration();
			++nFixL;
		}
		else if(sEvt) {
			if(!sEvt->isTrackLoss()) {
				meanSaccDurL += sEvt->duration();
				++nSaccL;
			}
			else ++nBlinkL;
		}
		ePtr = ePtr->next();
	}

//	ePtr = _rightEvents;
//	ePtr = _leftEvents;
	if (hasRightData())
		ePtr = _rightEvents;
	else
		ePtr = _leftEvents;
	while(ePtr) {                                //compute right mean fixn & sacc durations
		const FixationEvent *fEvt = dynamic_cast<FixationEvent*>(ePtr);
		const SaccadeEvent *sEvt = dynamic_cast<SaccadeEvent*>(ePtr);

		if(fEvt) {
			meanFixDurR += fEvt->duration();
			++nFixR;
			if(_disp2Time > 0 && fixIdx < 0 && fEvt->startTime() >= (_disp2Time-10)) {
				SaccadeEvent *tmp = fEvt->leadSaccade();
				fixIdx = fEvt->index();
				startFixPostDispChange = fEvt->startTime();
				endFixPostDispChange = fEvt->endTime();
				if(tmp) {
					saccStart = tmp->startTime();
					saccEnd = tmp->endTime();
					int aaa = fEvt->pixelLocation().x;
					int bbb = _dispChangeCoords.x;
					dcOK = (saccStart <= _disp2Time && saccEnd >= _disp2Time );
					dcWithin5 = ((saccStart) <= _disp2Time && (saccEnd +5) >= _disp2Time);
					dcWithin10 = ((saccStart) <= _disp2Time && (saccEnd +10) >= _disp2Time);
					if(saccEnd - saccStart > 50){
						hasLongSacc = true;
					}
					dcLeadSaccStart = saccStart;
					dcLeadSaccEnd = saccEnd;

				/*	if(!dcOK) {
							_flags.set(TRIAL_DELETED, true);
					}*/

					EyeEvent *p = tmp->previous(true);
					while(p) {
						const FixationEvent *fPrev = dynamic_cast<FixationEvent*>(p);
						const SaccadeEvent *sPrev = dynamic_cast<SaccadeEvent*>(p);
						if(fPrev) {
							//if(fPrev->endTime() < _disp2Time ) {
								endFixPreDispChange = fPrev->endTime();
								break;
							//}
							//else
							//	p = p->previous(true);
						}
						else if(sPrev) 
							p = p->previous(true);
					}
				}
			}
		}
		else if(sEvt) {
			if(sEvt->duration() > 50){
				if(endFixPostDispChange == (sEvt->startTime() - 1)){
					hasLongSacc = true;
				}
			}
			if(!sEvt->isTrackLoss()) {
				meanSaccDurR += sEvt->duration();
				/*if(sEvt->duration() > 50){
					int thesacc = sEvt->duration();
					hasLongSacc = true;
				}*/
				++nSaccR;
			}
			else {
				/*if(sEvt->duration() > 50){
					hasLongSacc = true;
				}*/
				++nBlinkR;
			}
		}
		ePtr = ePtr->next();
	}

	meanFixDurR /= nFixR;
	meanFixDurL /= nFixL;
	meanSaccDurR /= nSaccR;
	meanSaccDurL /= nSaccL;

	ePtr = _leftEvents;
	while(ePtr) {                                //compute left variances
		const FixationEvent *fEvt = dynamic_cast<FixationEvent*>(ePtr);
		const SaccadeEvent *sEvt = dynamic_cast<SaccadeEvent*>(ePtr);

		if(fEvt) {
			float diff = fEvt->duration() - meanFixDurL;
			varFixDurL += diff * diff;
		}
		else if(sEvt && !sEvt->isTrackLoss()) {
			float diff = sEvt->duration() - meanSaccDurL;
			varSaccDurL += diff * diff;
		}
		ePtr = ePtr->next();
	}

	ePtr = _rightEvents;
	while(ePtr) {                                //compute right variances
		const FixationEvent *fEvt = dynamic_cast<FixationEvent*>(ePtr);
		const SaccadeEvent *sEvt = dynamic_cast<SaccadeEvent*>(ePtr);

		if(fEvt) {
			float diff = fEvt->duration() - meanFixDurR;
			varFixDurR += diff * diff;
		}
		else if(sEvt && !sEvt->isTrackLoss()) {
			float diff = sEvt->duration() - meanSaccDurR;
			varSaccDurR += diff * diff;
		}
		ePtr = ePtr->next();
	}

	if(nFixR > 1) {
		varFixDurR /= (nFixR-1);
		varSaccDurR /= (nSaccR-1);
	}
	if(nFixL > 1) {
		varFixDurL /= (nFixL-1);
		varSaccDurL /= (nSaccL-1);
	}

	if(tabbed) {
		os<<index()<<'\t'<<_trialId.idString()<<'\t';

		os<<nFixR<<'\t';
		if(nFixR > 0) {
			os<<meanFixDurR<<'\t'<<sqrt(varFixDurR)<<'\t';
			os<<meanSaccDurR<<'\t'<<sqrt(varSaccDurR)<<'\t';
		}
		else os<<"0.0\t0.0\t0.0\t0.0\t";
		os<<nBlinkR<<'\t';

		os<<nFixL<<'\t';
		if(nFixL > 0) {
			os<<meanFixDurL<<'\t'<<sqrt(varFixDurL)<<'\t';
			os<<meanSaccDurL<<'\t'<<sqrt(varSaccDurL)<<'\t';
		}
		else os<<"0.0\t0.0\t0.0\t0.0\t";
		os<<nBlinkL<<'\t';

		if(_qstnAnswer > 0) 
			os<<(_endButton == _qstnAnswer ? "1" : "-1")<<'\t';
		else os<<"0"<<'\t';

		os<<endFixPreDispChange<<'\t';
		os<<_disp2Time<<'\t';
		os<<startFixPostDispChange<<'\t';

		if (_disp2Time > 0){
			os<<dcOK<<'\t';//<<std::endl;
			os<<dcWithin5<<'\t';
			os<<dcWithin10<<'\t';
		}
		else {
			os<<"-1"<<'\t';//<<std::endl;
			os<<"-1"<<'\t';
			os<<"-1"<<'\t';
		}

		os<<!(this->testFlag(BLINK_CHANGE_ERROR))<<'\t';
		os<<!(this->testFlag(J_HOOK_DISP_CHANGE))<<std::endl;


	}
	else {
		os<<"Trial ID:                       "<<_trialId.idString()<<"\r\n";
		os<<"Trial Type:                     ";
		if(_disp3Time > 0) os<<"FAST PRIME\r\n";
		else if(_disp2Time > 0) os<<"PREVIEW\r\n";
		else os<<"EYE TRACKING\r\n";

		if(_disp2Time > 0) {
			os<<"First Display Change\r\n";
			if( _previewBdry )
				os<<"    Boundary Character:         "<<_previewBdry<<"\r\n";
			os<<"    Fixation After Boundary:    "<<fixIdx<<"\r\n";
			os<<"    Leading Saccade Times:      "<<saccStart<<" -- "<<saccEnd<<"\r\n";
			os<<"    Display Change Time:        "<<_disp2Time<<"\r\n";
			os<<"    Change During Saccade?      "<<(dcOK ? "YES" : "NO")<<"\r\n";
		}
		if(_disp3Time > 0) {
			os<<"Second Diplay Change\r\n";
			os<<"    Display Change Time:        "<<_disp3Time<<"\r\n";
		}

		if(nFixR > 0) {
			os<<"Right Eye Fixations             "<<nFixR<<"\r\n";
			os<<"                    Mean:       "<<meanFixDurR<<"ms\r\n";
			os<<"                    Std. Dev.:  "<<sqrt(varFixDurR)<<"ms\r\n";
			os<<"          Saccade   Mean:       "<<meanSaccDurR<<"ms\r\n";
			os<<"                    Std. Dev.:  "<<sqrt(varSaccDurR)<<"ms\r\n";
			os<<"          Number Blinks:        "<<nBlinkR<<"\r\n";
		}
		if(nFixL > 0) {
			os<<"Left Eye Fixations              "<<nFixL<<"\r\n";
			os<<"                    Mean:       "<<meanFixDurL<<"ms\r\n";
			os<<"                    Std. Dev.:  "<<sqrt(varFixDurL)<<"ms\r\n";
			os<<"         Saccade    Mean:       "<<meanSaccDurL<<"ms\r\n";
			os<<"                    Std. Dev.:  "<<sqrt(varSaccDurL)<<"ms\r\n";
			os<<"         Number Blinks:         "<<nBlinkL<<"\r\n";
		}
		if(_qstnAnswer > 0)
			os<<"Subject Response Is Correct?    "<<(_endButton == _qstnAnswer ? "YES" : "NO")<<"\r\n";
		os<<"___________________________________________\r\n\r\n";
	}

	os.flush();

}




void TrialData::setDisplayCoords(CRect r)
{
	_displayCoords = r;
}

unsigned int TrialData::getLastLeadStart(unsigned int curFixTime){
	std::vector<DCDATA>::iterator dc = _dcVector.begin();
	while(dc != _dcVector.end()){
		if(dc->time > curFixTime){
			return ((--dc)->leadStart);
		}
		else {
				dc++;
			}
		}
	return (_dcVector.back().leadStart);
}

unsigned int TrialData::getLastLeadEnd(unsigned int curFixTime){
	std::vector<DCDATA>::iterator dc = _dcVector.begin();
	while(dc != _dcVector.end()){
		if(dc->time > curFixTime){
			return ((--dc)->leadEnd);
		}
		else {
				dc++;
			}
		}
	return (_dcVector.back().leadEnd);
}

unsigned int TrialData::getLastDCTime(unsigned int curFixTime){
	std::vector<DCDATA>::iterator dc = _dcVector.begin();
	while(dc != _dcVector.end()){
		if(dc->time >= curFixTime){
			return ((--dc)->time);
		}
		else {
				dc++;
			}
		}
	//return (_dcVector.front().time);
	return(_dcVector.back().time);
}

int TrialData::dcTime(){
	if(_disp2Time)
		return _disp2Time;
	else return 0;
}

const imagePos &imagePos::operator=(const imagePos &ip)
{
	fileName = ip.fileName;
	upperLeft = ip.upperLeft;
	return(*this);
}

imagePos::imagePos()
: fileName(""), upperLeft(-1,-1)
{}

imagePos::imagePos(std::string fName, CPoint pt)
{
	fileName = fName;
	upperLeft = pt;
}

imagePos::~imagePos()
{}

/*const &dcData::operator=(const dcData &dc)
{
	loc = dc.loc;
	time = dc.time;
	return(*this);
}*/

dcData::dcData()
: loc(0,0), time(0), leadStart(0), leadEnd(0), fixTime(0)
{}

dcData::dcData(CPoint pt, unsigned int t)
{
	loc = pt;
	time = t; 

}

dcData::dcData(CPoint pt, unsigned int t, unsigned int ft,unsigned int lt, unsigned int et)
{
	loc = pt;
	time = t; 
	fixTime = ft;
	leadStart = lt;
	leadEnd = et;
}

void dcData::read(std::istream &is, int versionNum){
	char* _text;
	CRect _boundary;
	CPoint _loc;
	int sz = sizeof(CRect) + sizeof(CPoint) + sizeof(short);
	int szLoc = sizeof(CPoint) + sizeof(short) + sizeof(bool);
	std::ios::pos_type pos;
	short len = 1;

	if(versionNum > 2)
		sz +=  sizeof(bool);
	is.read(reinterpret_cast<char*>(&_boundary), sz); //read region data
	is.read(reinterpret_cast<char*>(&_loc), szLoc);

	pos = is.tellg();                            //mark current position in stream
	while(is.get() != '\0') ++len;               //find the end (length) of text string

	is.seekg(pos);                               //rewind to beginning of text
	_text = new char[len];
	is.read(_text, len);                         //read in text



	this->loc = _loc;
	this->time = _boundary.left;
	this->leadEnd = _boundary.top;
	this->leadStart = _boundary.right;
	this->fixTime = _boundary.bottom;
}

void dcData::write(std::ostream &os) const{
	char _text[5] = "test";
	CRect _times = CRect(this->time,this->leadEnd,this->leadStart,this->fixTime);
	CPoint _loc = this->loc;
	int sz = sizeof(CRect) + sizeof(CPoint) + sizeof(short) + sizeof(bool);
	int szLoc = sizeof(CPoint) + sizeof(short) + sizeof(bool);
	short len = 0;

	while(_text[len] != '\0') ++len;             //compute length of string
                                                 //write region data
	os.write(reinterpret_cast<const char*>(&_times), sz);
	os.write(reinterpret_cast<const char*>(&_loc), szLoc);
	os.write(_text, len+1);
}


dcData TrialData::getDCData(unsigned int eTime){
	if(eTime<=0){
		return( dcData(CPoint(0,0),0,0,0,0));
	}
	std::vector<DCDATA>::iterator dc = _dcVector.begin();
	bool first = true;
	while(dc != _dcVector.end()){
		if(dc->time >= eTime){
			if(first){
				return( dcData(CPoint(0,0),0,0,0,0));
			}
			else{
				--dc;
				return (dcData(CPoint(dc->loc.x,dc->loc.y),dc->time,dc->fixTime,dc->leadStart, dc->leadEnd));
			}
		}
		else {
				dc++;
				first = false;
			}
	}
	//return (_dcVector.front().time);
	dcData d = _dcVector.back();
	//return(dcData(CPoint(0,0),0,0,0,0));
	return(dcData(CPoint(d.loc.x,d.loc.y),d.time,d.fixTime,d.leadStart, d.leadEnd));
}


dcData::~dcData()
{}